package org.processmining.acceptingpetrinet.algorithms;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetFactory;
import org.processmining.acceptingpetrinet.parameters.UndoubleAcceptingPetriNetParameters;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.semantics.petrinet.Marking;

public class UndoubleAcceptingPetriNetAlgorithm {

	public AcceptingPetriNet apply(PluginContext context, AcceptingPetriNet apn, UndoubleAcceptingPetriNetParameters parameters) {
		Petrinet net = apn.getNet();
		Set<String> duplicateLabels = new HashSet<String>();
		Set<String> labels = new HashSet<String>();
		/*
		 * Check whether we have duplicate labels.
		 */
		for (Transition transition : net.getTransitions()) {
			if (!transition.isInvisible()) {
				if (labels.contains(transition.getLabel())) {
					duplicateLabels.add(transition.getLabel());
				} else {
					labels.add(transition.getLabel());
				}
			}
		}
		/*
		 * Nope. Original net will do.
		 */
		if (duplicateLabels.isEmpty()) {
			return apn;
		}

		/*
		 * Create a new net with duplicate labels 'unduplicated'.
		 */
		Petrinet newNet = PetrinetFactory.newPetrinet(net.getLabel());
		/*
		 * Copy all places.
		 */
		Map<Place, Place> placeMap = new HashMap<Place, Place>(net.getPlaces().size());
		for (Place place : net.getPlaces()) {
			placeMap.put(place, newNet.addPlace(place.getLabel()));
		}
		Map<Transition, Transition> transitionMap = new HashMap<Transition, Transition>(net.getTransitions().size());
		/*
		 * Every occurrence of a transition with duplicate label is replaced by
		 * a non-shared fragment and a shared fragment. The shared fragment
		 * contains a start place, a transition with the unique label, and an
		 * end place. The non-shared fragment contains a start transition, an
		 * intermediate place, and an end transition. The start transition puts
		 * a token in the start place, and the end transition needs one from the
		 * end place.
		 */
		Map<String, Place> startPlaceMap = new HashMap<String, Place>(net.getTransitions().size());
		Map<String, Place> endPlaceMap = new HashMap<String, Place>(net.getTransitions().size());
		Map<Transition, Transition> startTransitionMap = new HashMap<Transition, Transition>(net.getTransitions()
				.size());
		Map<Transition, Transition> endTransitionMap = new HashMap<Transition, Transition>(net.getTransitions().size());
		/*
		 * Transfer all transitions.
		 */
		for (Transition transition : net.getTransitions()) {
			String label = transition.getLabel();
			if (transition.isInvisible() || !duplicateLabels.contains(label)) {
				/*
				 * No duplicate label, Just copy.
				 */
				transitionMap.put(transition, newNet.addTransition(label));
				transitionMap.get(transition).setInvisible(transition.isInvisible());
			} else {
				/*
				 * Duplicate label. Create fragment(s).
				 */
				if (!startPlaceMap.containsKey(label)) {
					/*
					 * No shared fragment yet, create it now.
					 */
					Transition doTransition = newNet.addTransition(label);
					startPlaceMap.put(label, newNet.addPlace("|p> " + label));
					endPlaceMap.put(label, newNet.addPlace("[p] " + label));
					newNet.addArc(startPlaceMap.get(label), doTransition);
					newNet.addArc(doTransition, endPlaceMap.get(label));
				}
				/*
				 * Create non-shared fragment.
				 */
				Transition startTransition = newNet.addTransition("|t> " + label);
				startTransition.setInvisible(true);
				startTransitionMap.put(transition, startTransition);
				newNet.addArc(startTransition, startPlaceMap.get(label));
				Transition endTransition = newNet.addTransition("[t] " + label);
				endTransition.setInvisible(true);
				endTransitionMap.put(transition, endTransition);
				newNet.addArc(endPlaceMap.get(label), endTransition);
				Place doPlace = newNet.addPlace(label);
				/*
				 * Connect both fragments.
				 */
				newNet.addArc(startTransition, doPlace);
				newNet.addArc(doPlace, endTransition);
			}
		}
		/*
		 * Transfer all arcs.
		 */
		for (PetrinetEdge<?, ?> edge : net.getEdges()) {
			if (edge instanceof Arc) {
				Arc arc = (Arc) edge;
				if (arc.getSource() instanceof Place) {
					Place place = placeMap.get(arc.getSource());
					Transition transition = transitionMap.get(arc.getTarget());
					if (transition == null) {
						/*
						 * Transition has duplicate label. Use startTransition
						 * instead.
						 */
						transition = startTransitionMap.get(arc.getTarget());
					}
					newNet.addArc(place, transition);
				} else {
					Transition transition = transitionMap.get(arc.getSource());
					if (transition == null) {
						/*
						 * Transition has duplicate label. Use endTransition
						 * instead.
						 */
						transition = endTransitionMap.get(arc.getSource());
					}
					Place place = placeMap.get(arc.getTarget());
					newNet.addArc(transition, place);
				}
			}
		}
		/*
		 * Create new accepting Petri net.
		 */
		AcceptingPetriNet newApn = AcceptingPetriNetFactory.createAcceptingPetriNet();
		newApn.init(newNet);
		
		/*
		 * Set initial marking.
		 */
		newApn.setInitialMarking(apply(apn.getInitialMarking(), placeMap));
		
		/*
		 * Set final markings.
		 */
		Set<Marking> finalMarkings = new HashSet<Marking>();
		for (Marking finalMarking: apn.getFinalMarkings()) {
			finalMarkings.add(apply(finalMarking, placeMap));
		}
		newApn.setFinalMarkings(finalMarkings);
		/*
		 * Return new accepting Petri net.
		 */
		return newApn;
	}

	private Marking apply(Marking marking, Map<Place, Place> map) {
		Marking appliedMarking = new Marking();
		for (Place place : marking.baseSet()) {
			appliedMarking.add(map.get(place), marking.occurrences(place));
		}
		return appliedMarking;
	}
}
