package org.processmining.pnanalysis.metrics.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.pnanalysis.metrics.PetriNetMetric;

public class PetriNetStructurednessMetric implements PetriNetMetric {

	public final static String NAME = "Structuredness metric";

	private Petrinet net;
	private Map<Transition, Double> metrics;
	private StringBuffer buffer;
	
	public double compute(PluginContext context, Petrinet net, Marking marking) {
		Petrinet clonedNet = PetrinetFactory.clonePetrinet(net);
		Set<Place> sourcePlaces = new HashSet<Place>();
		Set<Place> sinkPlaces = new HashSet<Place>();
		Set<Transition> sourceTransitions = new HashSet<Transition>();
		Set<Transition> sinkTransitions = new HashSet<Transition>();
		for (Place place : clonedNet.getPlaces()) {
			if (clonedNet.getInEdges(place).isEmpty()) {
				sourcePlaces.add(place);
			}
			if (clonedNet.getOutEdges(place).isEmpty()) {
				sinkPlaces.add(place);
			}
		}
		for (Transition transition : clonedNet.getTransitions()) {
			if (clonedNet.getInEdges(transition).isEmpty()) {
				sourceTransitions.add(transition);
			}
			if (clonedNet.getOutEdges(transition).isEmpty()) {
				sinkTransitions.add(transition);
			}
		}
		buffer = new StringBuffer();
		if (sourcePlaces.size() == 1 && sinkPlaces.isEmpty() && sourceTransitions.isEmpty() && sinkTransitions.size() == 1) {
			Place sinkPlace = clonedNet.addPlace("sink");
			clonedNet.addArc(sinkTransitions.iterator().next(), sinkPlace);
			sinkPlaces.add(sinkPlace);
			sinkTransitions = new HashSet<Transition>();
		}
		if (sourcePlaces.size() == 1 && sinkPlaces.size() == 1 && sourceTransitions.isEmpty() && sinkTransitions.isEmpty()) {
			this.net = clonedNet;
			metrics = new HashMap<Transition, Double>();
			for (Transition transition : clonedNet.getTransitions()) {
				metrics.put(transition, 1.0);
			}
			return getMetric(clonedNet);
		}
		buffer.append("Net is not a workflow net.");
		return -1.0;
	}

	public String getName() {
		return NAME;
	}

	private double getMetric(Petrinet net) {
		while (net.getTransitions().size() > 1) {
			if (foldSequence()) {
				continue;
			}
			if (foldChoice()) {
				continue;
			}
			if (foldWhile()) {
				continue;
			}
			if (foldMarkedGraph()) {
				continue;
			}
			if (foldStateMachine()) {
				continue;
			}
			if (foldWellStructured()) {
				continue;
			}
			if (foldUnstructured()) {
				continue;
			}
		}
		return metrics.get(net.getTransitions().iterator().next());
	}

	private boolean foldSequence() {
		for (Place place : net.getPlaces()) {
			if (net.getInEdges(place).size() != 1 || net.getOutEdges(place).size() != 1) {
				continue;
			}
			Transition sourceTransition = (Transition) net.getInEdges(place).iterator().next().getSource();
			Transition sinkTransition = (Transition) net.getOutEdges(place).iterator().next().getTarget();
			if (sourceTransition == sinkTransition || net.getOutEdges(sourceTransition).size() > 1
					|| net.getInEdges(sinkTransition).size() > 1) {
				continue;
			}
			buffer.append("SEQUENCE " + place.getLabel() + "\n");
			net.removePlace(place);
			buffer.append("Removed place " + place.getLabel() + "\n");
			for (PetrinetEdge<?, ?> edge : net.getOutEdges(sinkTransition)) {
				net.addArc(sourceTransition, (Place) edge.getTarget());
				buffer.append("Added arc from " + sourceTransition.getLabel() + " to " + edge.getTarget().getLabel()
						+ "\n");
			}
			net.removeTransition(sinkTransition);
			buffer.append("Removed transition " + sinkTransition.getLabel() + "\n");
			metrics.put(sourceTransition, metrics.get(sourceTransition) + metrics.get(sinkTransition));
			buffer.append("Sum = " + metrics.get(sourceTransition) + "\n");
			buffer.append("Metric = " + metrics.get(sourceTransition) + "\n");
			return true;
		}
		return false;
	}

	private boolean foldChoice() {
		for (Transition transition : net.getTransitions()) {
			if (net.getInEdges(transition).size() != 1 || net.getOutEdges(transition).size() != 1) {
				continue;
			}
			Place sourcePlace = (Place) net.getInEdges(transition).iterator().next().getSource();
			Place sinkPlace = (Place) net.getOutEdges(transition).iterator().next().getTarget();
			if (sourcePlace == sinkPlace) {
				continue;
			}
			Set<Transition> targetTransitions = new HashSet<Transition>();
			Set<Transition> sourceTransitions = new HashSet<Transition>();
			for (PetrinetEdge<?, ?> edge : net.getOutEdges(sourcePlace)) {
				targetTransitions.add((Transition) edge.getTarget());
			}
			for (PetrinetEdge<?, ?> edge : net.getInEdges(sinkPlace)) {
				sourceTransitions.add((Transition) edge.getSource());
			}
			if (targetTransitions.size() < 2) {
				continue;
			}
			if (!targetTransitions.equals(sourceTransitions)) {
				continue;
			}
			buffer.append("CHOICE " + transition.getLabel() + "\n");
			double metric = 0.0;
			for (Transition targetTransition : targetTransitions) {
				metric += metrics.get(targetTransition);
				if (targetTransition != transition) {
					net.removeTransition(targetTransition);
					buffer.append("Removed transition " + targetTransition.getLabel() + "\n");
				}
			}
			buffer.append("Sum = " + metric + "\n");
			metric *= 1.5;
			metrics.put(transition, metric);
			buffer.append("Metric = " + metrics.get(transition) + "\n");
			return true;
		}
		return false;
	}

	private boolean foldWhile() {
		for (Place place : net.getPlaces()) {
			if (net.getInEdges(place).size() != 2 || net.getOutEdges(place).size() != 2) {
				continue;
			}
			Set<Transition> sourceTransitions = new HashSet<Transition>();
			Iterator<PetrinetEdge<?, ?>> sourceIterator = net.getInEdges(place).iterator();
			sourceTransitions.add((Transition) sourceIterator.next().getSource());
			sourceTransitions.add((Transition) sourceIterator.next().getSource());
			Set<Transition> targetTransitions = new HashSet<Transition>();
			Iterator<PetrinetEdge<?, ?>> targetIterator = net.getOutEdges(place).iterator();
			targetTransitions.add((Transition) targetIterator.next().getTarget());
			targetTransitions.add((Transition) targetIterator.next().getTarget());
			Set<Transition> sharedTransitions = new HashSet<Transition>(sourceTransitions);
			sharedTransitions.retainAll(targetTransitions);
			if (sharedTransitions.size() != 1) {
				continue;
			}
			sourceTransitions.removeAll(sharedTransitions);
			targetTransitions.removeAll(sharedTransitions);
			Transition sourceTransition = sourceTransitions.iterator().next();
			Transition sharedTransition = sharedTransitions.iterator().next();
			Transition sinkTransition = targetTransitions.iterator().next();
			if (net.getInEdges(sharedTransition).size() != 1 || net.getOutEdges(sharedTransition).size() != 1) {
				continue;
			}
			if (net.getInEdges(sinkTransition).size() != 1 || net.getOutEdges(sourceTransition).size() != 1) {
				continue;
			}
			buffer.append("WHILE " + place.getLabel() + "\n");
			net.removePlace(place);
			buffer.append("Removed place " + place.getLabel() + "\n");
			metrics.put(sourceTransition,
					metrics.get(sourceTransition) + metrics.get(sinkTransition) + 2.0 * metrics.get(sharedTransition));
			net.removeTransition(sharedTransition);
			buffer.append("Removed transition " + sharedTransition.getLabel() + "\n");
			for (PetrinetEdge<?, ?> edge : net.getOutEdges(sinkTransition)) {
				net.addArc(sourceTransition, (Place) edge.getTarget());
				buffer.append("Added arc from " + sourceTransition.getLabel() + " to " + edge.getTarget().getLabel()
						+ "\n");
			}
			net.removeTransition(sinkTransition);
			buffer.append("Removed transition " + sinkTransition.getLabel() + "\n");
			buffer.append("Sum = " + metrics.get(sourceTransition) + "\n");
			buffer.append("Metric = " + metrics.get(sourceTransition) + "\n");
			return true;
		}
		return false;
	}

	private boolean foldMarkedGraph() {
		Set<PetrinetNode> bestNodes = new HashSet<PetrinetNode>();
		Transition bestSourceTransition = null;
		Transition bestSinkTransition = null;
		for (Transition sourceTransition : net.getTransitions()) {
			for (Transition sinkTransition : net.getTransitions()) {
				Set<PetrinetNode> forwardNodes = forward(sourceTransition, sinkTransition);
				Set<PetrinetNode> backwardNodes = backward(sinkTransition, sourceTransition);
				int nofTransitions = 0;
				for (PetrinetNode forwardNode : forwardNodes) {
					if (forwardNode instanceof Transition) {
						nofTransitions++;
					}
				}
				if (nofTransitions > 1 && forwardNodes.equals(backwardNodes)) {
					boolean isMarkedGraph = true;
					for (PetrinetNode node : forwardNodes) {
						if (node instanceof Place
								&& (net.getInEdges(node).size() != 1 || net.getOutEdges(node).size() != 1)) {
							isMarkedGraph = false;
						}
					}
					if (isMarkedGraph) {
						if (forwardNodes.size() > bestNodes.size()) {
							bestNodes = forwardNodes;
							bestSourceTransition = sourceTransition;
							bestSinkTransition = sinkTransition;
						}
					}
				}
			}
		}
		if (bestNodes.size() > 0) {
			buffer.append("MARKED GRAPH " + bestSourceTransition.getLabel() + " -> " + bestSinkTransition.getLabel()
					+ "\n");
			double metric = 0.0;
			int diffPost = 0;
			int diffPre = 0;
			for (PetrinetNode node : bestNodes) {
				diffPost += net.getOutEdges(node).size() > 1 ? 1 : 0;
				diffPre += net.getInEdges(node).size() > 1 ? 1 : 0;
				if (node instanceof Transition) {
					metric += 2.0 * metrics.get(node);
				}
			}
			int diff = Math.abs(diffPre - diffPost) + 1;
			buffer.append("Sum = " + metric + "\n");
			buffer.append("DiffT = " + diff + "\n");
			for (PetrinetEdge<?, ?> edge : net.getOutEdges(bestSinkTransition)) {
				net.addArc(bestSourceTransition, (Place) edge.getTarget());
				buffer.append("Added arc from " + bestSourceTransition.getLabel() + " to "
						+ edge.getTarget().getLabel() + "\n");
			}
			for (PetrinetNode node : bestNodes) {
				if (node != bestSourceTransition) {
					if (node instanceof Place) {
						net.removePlace((Place) node);
						buffer.append("Removed place " + node.getLabel() + "\n");
					} else {
						net.removeTransition((Transition) node);
						buffer.append("Removed transition " + node.getLabel() + "\n");
					}
				}
			}
			metrics.put(bestSourceTransition, metric * diff);
			buffer.append("Metric = " + metrics.get(bestSourceTransition) + "\n");
			return true;
		}
		return false;
	}

	private boolean foldStateMachine() {
		Set<PetrinetNode> bestNodes = new HashSet<PetrinetNode>();
		Place bestSourcePlace = null;
		Place bestSinkPlace = null;
		for (Place sourcePlace : net.getPlaces()) {
			for (Place sinkPlace : net.getPlaces()) {
				Set<PetrinetNode> forwardNodes = forward(sourcePlace, sinkPlace);
				Set<PetrinetNode> backwardNodes = backward(sinkPlace, sourcePlace);
				int nofTransitions = 0;
				for (PetrinetNode forwardNode : forwardNodes) {
					if (forwardNode instanceof Transition) {
						nofTransitions++;
					}
				}
				if (nofTransitions > 1 && forwardNodes.equals(backwardNodes)) {
					boolean isStateMachine = true;
					for (PetrinetNode node : forwardNodes) {
						if (node instanceof Transition
								&& (net.getInEdges(node).size() != 1 || net.getOutEdges(node).size() != 1)) {
							isStateMachine = false;
						}
					}
					if (isStateMachine) {
						if (forwardNodes.size() > bestNodes.size()) {
							bestNodes = forwardNodes;
							bestSourcePlace = sourcePlace;
							bestSinkPlace = sinkPlace;
						}
					}
				}
			}
		}
		if (bestNodes.size() > 0) {
			buffer.append("STATE MACHINE " + bestSourcePlace.getLabel() + " -> " + bestSinkPlace.getLabel() + "\n");
			double metric = 0.0;
			int diffPost = 0;
			int diffPre = 0;
			for (PetrinetNode node : bestNodes) {
				diffPost += net.getOutEdges(node).size() > 1 ? 1 : 0;
				diffPre += net.getInEdges(node).size() > 1 ? 1 : 0;
				if (node instanceof Transition) {
					metric += 2.0 * metrics.get(node);
				}
			}
			int diff = Math.abs(diffPre - diffPost) + 1;
			buffer.append("Sum = " + metric + "\n");
			buffer.append("DiffP = " + diff + "\n");
			Transition survivorTransition = (Transition) net.getOutEdges(bestSourcePlace).iterator().next().getTarget();
			net.addArc(survivorTransition, bestSinkPlace);
			buffer.append("Added arc from " + survivorTransition.getLabel() + " to " + bestSinkPlace.getLabel() + "\n");
			for (PetrinetNode node : bestNodes) {
				if (node != bestSourcePlace && node != bestSinkPlace && node != survivorTransition) {
					if (node instanceof Place) {
						net.removePlace((Place) node);
						buffer.append("Removed place " + node.getLabel() + "\n");
					} else {
						net.removeTransition((Transition) node);
						buffer.append("Removed transition " + node.getLabel() + "\n");
					}
				}
			}
			metrics.put(survivorTransition, metric * diff);
			buffer.append("Metric = " + metrics.get(survivorTransition) + "\n");
			return true;
		}
		return false;
	}

	private boolean foldWellStructured() {
		Set<PetrinetNode> bestNodes = new HashSet<PetrinetNode>();
		PetrinetNode bestSourceNode = null;
		PetrinetNode bestSinkNode = null;
		for (PetrinetNode sourceNode : net.getPlaces()) {
			for (PetrinetNode sinkNode : net.getPlaces()) {
				Set<PetrinetNode> forwardNodes = forward(sourceNode, sinkNode);
				Set<PetrinetNode> backwardNodes = backward(sinkNode, sourceNode);
				int nofTransitions = 0;
				for (PetrinetNode forwardNode : forwardNodes) {
					if (forwardNode instanceof Transition) {
						nofTransitions++;
					}
				}
				if (nofTransitions > 1 && forwardNodes.equals(backwardNodes)) {
					boolean isWellStructured = true;
					for (PetrinetNode node : forwardNodes) {
						if (node != sinkNode && node instanceof Place) {
							for (PetrinetEdge<?, ?> outEdge : net.getOutEdges(node)) {
								if (net.getOutEdges(node).size() > 1 && net.getInEdges(outEdge.getTarget()).size() > 1) {
									isWellStructured = false;
								}
								if (net.getInEdges(node).size() > 1 && net.getInEdges(outEdge.getTarget()).size() > 1) {
									isWellStructured = false;
								}
								if (forward(outEdge.getTarget(), sinkNode).contains(node)) {
									isWellStructured = false;
								}
							}
						}
					}
					if (isWellStructured) {
						if (forwardNodes.size() > bestNodes.size()) {
							bestNodes = forwardNodes;
							bestSourceNode = sourceNode;
							bestSinkNode = sinkNode;
						}
					}
				}
			}
		}
		if (bestNodes.size() > 0) {
			buffer.append("WELL STRUCTURED " + bestSourceNode.getLabel() + " -> " + bestSinkNode.getLabel() + "\n");
			double metric = 0.0;
			int diffPlacePost = 0;
			int diffPlacePre = 0;
			int diffTransitionPost = 0;
			int diffTransitionPre = 0;
			for (PetrinetNode node : bestNodes) {
				if (node instanceof Place) {
					diffPlacePost += net.getOutEdges(node).size() > 1 ? 1 : 0;
					diffPlacePre += net.getInEdges(node).size() > 1 ? 1 : 0;
				} else {
					diffTransitionPost += net.getOutEdges(node).size() > 1 ? 1 : 0;
					diffTransitionPre += net.getInEdges(node).size() > 1 ? 1 : 0;
				}
				if (node instanceof Transition) {
					metric += 2.0 * metrics.get(node);
				}
			}
			int diffPlace = Math.abs(diffPlacePre - diffPlacePost) + 1;
			int diffTransition = Math.abs(diffTransitionPre - diffTransitionPost) + 1;
			buffer.append("Sum = " + metric + "\n");
			buffer.append("DiffP" + diffPlace + "\n");
			buffer.append("DiffT = " + diffTransition + "\n");
			Transition survivorTransition;
			if (bestSourceNode instanceof Place) {
				if (bestSinkNode instanceof Place) {
					survivorTransition = (Transition) net.getOutEdges(bestSourceNode).iterator().next().getTarget();
					net.addArc(survivorTransition, (Place) bestSinkNode);
					buffer.append("Added arc from " + survivorTransition.getLabel() + " to " + bestSinkNode.getLabel()
							+ "\n");
				} else {
					survivorTransition = (Transition) bestSinkNode;
					net.addArc((Place) bestSourceNode, survivorTransition);
					buffer.append("Added arc from " + bestSourceNode.getLabel() + " to "
							+ survivorTransition.getLabel() + "\n");
				}
			} else {
				if (bestSinkNode instanceof Place) {
					survivorTransition = (Transition) bestSourceNode;
					net.addArc(survivorTransition, (Place) bestSinkNode);
					buffer.append("Added arc from " + survivorTransition.getLabel() + " to " + bestSinkNode.getLabel()
							+ "\n");
				} else {
					survivorTransition = (Transition) bestSourceNode;
					for (PetrinetEdge<?, ?> outEdge : net.getOutEdges(bestSinkNode)) {
						net.addArc(survivorTransition, (Place) outEdge.getTarget());
						buffer.append("Added arc from " + survivorTransition.getLabel() + " to "
								+ outEdge.getTarget().getLabel() + "\n");
					}
					bestSinkNode = null;
				}
			}
			for (PetrinetNode node : bestNodes) {
				if (node != bestSourceNode && node != bestSinkNode && node != survivorTransition) {
					if (node instanceof Place) {
						net.removePlace((Place) node);
						buffer.append("Removed place " + node.getLabel() + "\n");
					} else {
						net.removeTransition((Transition) node);
						buffer.append("Removed transition " + node.getLabel() + "\n");
					}
				}
			}
			metrics.put(survivorTransition, metric * diffPlace * diffTransition);
			buffer.append("Metric = " + metrics.get(survivorTransition) + "\n");
			return true;
		}
		return false;
	}

	private boolean foldUnstructured() {
		Set<PetrinetNode> bestNodes = new HashSet<PetrinetNode>();
		PetrinetNode bestSourceNode = null;
		PetrinetNode bestSinkNode = null;
		for (PetrinetNode sourceNode : net.getPlaces()) {
			for (PetrinetNode sinkNode : net.getPlaces()) {
				Set<PetrinetNode> forwardNodes = forward(sourceNode, sinkNode);
				Set<PetrinetNode> backwardNodes = backward(sinkNode, sourceNode);
				int nofTransitions = 0;
				for (PetrinetNode forwardNode : forwardNodes) {
					if (forwardNode instanceof Transition) {
						nofTransitions++;
					}
				}
				if (nofTransitions > 1 && forwardNodes.equals(backwardNodes)) {
					if (bestNodes.size() == 0 || forwardNodes.size() < bestNodes.size()) {
						bestNodes = forwardNodes;
						bestSourceNode = sourceNode;
						bestSinkNode = sinkNode;
					}
				}
			}
		}
		if (bestNodes.size() > 0) {
			buffer.append("UNSTRUCTURED " + bestSourceNode.getLabel() + " -> " + bestSinkNode.getLabel() + "\n");
			double metric = 0.0;
			int diffPlacePost = 0;
			int diffPlacePre = 0;
			int diffTransitionPost = 0;
			int diffTransitionPre = 0;
			int nofArcs = 0;
			int nofPlaces = 0;
			int nofTransitions = 0;
			for (Place place : net.getPlaces()) {
				if (bestNodes.contains(place)) {
					nofPlaces++;
				}
			}
			for (Transition transition : net.getTransitions()) {
				if (bestNodes.contains(transition)) {
					nofTransitions++;
				}
			}
			for (PetrinetEdge<?, ?> edge : net.getEdges()) {
				if (bestNodes.contains(edge.getSource()) && bestNodes.contains(edge.getTarget())) {
					nofArcs++;
				}
			}
			int weight = 5 * (Math.abs(nofArcs - (nofPlaces + nofTransitions)) + 1);
			for (PetrinetNode node : bestNodes) {
				if (node instanceof Place) {
					if (net.getOutEdges(node).size() > 1) {
						buffer.append("Out place " + node.getLabel() + "\n");
					}
					if (net.getInEdges(node).size() > 1) {
						buffer.append("In place " + node.getLabel() + "\n");
					}
					diffPlacePost += net.getOutEdges(node).size() > 1 ? 1 : 0;
					diffPlacePre += net.getInEdges(node).size() > 1 ? 1 : 0;
				} else {
					if (net.getOutEdges(node).size() > 1) {
						buffer.append("Out transition " + node.getLabel() + "\n");
					}
					if (net.getInEdges(node).size() > 1) {
						buffer.append("In transition " + node.getLabel() + "\n");
					}
					diffTransitionPost += net.getOutEdges(node).size() > 1 ? 1 : 0;
					diffTransitionPre += net.getInEdges(node).size() > 1 ? 1 : 0;
				}
				if (node instanceof Transition) {
					metric += weight * metrics.get(node);
				}
			}
			int diffPlace = Math.abs(diffPlacePre - diffPlacePost) + 1;
			int diffTransition = Math.abs(diffTransitionPre - diffTransitionPost) + 1;
			buffer.append("|F| = " + nofArcs + "\n");
			buffer.append("|P U T| = " + (nofPlaces + nofTransitions) + "\n");
			buffer.append("Sum = " + metric + "\n");
			buffer.append("DiffP = " + diffPlace + "\n");
			buffer.append("DiffT = " + diffTransition + "\n");
			Transition survivorTransition;
			if (bestSourceNode instanceof Place) {
				if (bestSinkNode instanceof Place) {
					survivorTransition = (Transition) net.getOutEdges(bestSourceNode).iterator().next().getTarget();
					net.addArc(survivorTransition, (Place) bestSinkNode);
					buffer.append("Added arc from " + survivorTransition.getLabel() + " to " + bestSinkNode.getLabel()
							+ "\n");
				} else {
					survivorTransition = (Transition) bestSinkNode;
					net.addArc((Place) bestSourceNode, survivorTransition);
					buffer.append("Added arc from " + bestSourceNode.getLabel() + " to "
							+ survivorTransition.getLabel() + "\n");
				}
			} else {
				if (bestSinkNode instanceof Place) {
					survivorTransition = (Transition) bestSourceNode;
					net.addArc(survivorTransition, (Place) bestSinkNode);
					buffer.append("Added arc from " + survivorTransition.getLabel() + " to " + bestSinkNode.getLabel()
							+ "\n");
				} else {
					survivorTransition = (Transition) bestSourceNode;
					for (PetrinetEdge<?, ?> outEdge : net.getOutEdges(bestSinkNode)) {
						net.addArc(survivorTransition, (Place) outEdge.getTarget());
						buffer.append("Added arc from " + survivorTransition.getLabel() + " to "
								+ outEdge.getTarget().getLabel() + "\n");
					}
					bestSinkNode = null;
				}
			}
			for (PetrinetNode node : bestNodes) {
				if (node != bestSourceNode && node != bestSinkNode && node != survivorTransition) {
					if (node instanceof Place) {
						net.removePlace((Place) node);
						buffer.append("Removed place " + node.getLabel() + "\n");
					} else {
						net.removeTransition((Transition) node);
						buffer.append("Removed transition " + node.getLabel() + "\n");
					}
				}
			}
			metrics.put(survivorTransition, metric * diffPlace * diffTransition);
			buffer.append("Metric = " + metrics.get(survivorTransition) + "\n");
			return true;
		}
		return false;
	}

	private Set<PetrinetNode> forward(PetrinetNode sourceNode, PetrinetNode sinkNode) {
		List<PetrinetNode> toDo = new ArrayList<PetrinetNode>();
		toDo.add(sourceNode);
		Set<PetrinetNode> nodes = new HashSet<PetrinetNode>();
		nodes.add(sourceNode);
		nodes.add(sinkNode);
		while (!toDo.isEmpty()) {
			PetrinetNode node = toDo.remove(0);
			nodes.add(node);
			for (PetrinetEdge<?, ?> edge : net.getOutEdges(node)) {
				if (!nodes.contains(edge.getTarget())) {
					toDo.add(edge.getTarget());
					nodes.add(edge.getTarget());
				}
			}
		}
		return nodes;
	}

	private Set<PetrinetNode> backward(PetrinetNode sinkNode, PetrinetNode sourceNode) {
		List<PetrinetNode> toDo = new ArrayList<PetrinetNode>();
		toDo.add(sinkNode);
		Set<PetrinetNode> nodes = new HashSet<PetrinetNode>();
		nodes.add(sinkNode);
		nodes.add(sourceNode);
		while (!toDo.isEmpty()) {
			PetrinetNode node = toDo.remove(0);
			nodes.add(node);
			for (PetrinetEdge<?, ?> edge : net.getInEdges(node)) {
				if (!nodes.contains(edge.getSource())) {
					toDo.add(edge.getSource());
					nodes.add(edge.getSource());
				}
			}
		}
		return nodes;
	}

	public StringBuffer getLog() {
		return buffer;
	}
}