package org.processmining.processtree;

import java.util.Iterator;
import java.util.List;

public interface Block extends Node {

	/**
	 * returns an immutable (nonempty) list outgoing edges.
	 * 
	 * @return
	 */
	public List<Edge> getOutgoingEdges();

	/**
	 * Adds an outgoing edge to the node
	 * 
	 * @param edge
	 */
	public void addOutgoingEdge(Edge edge);
	
	/**
	 * Removes and outgoing edge from the node
	 * 
	 * @param edge
	 */
	public void removeOutgoingEdge(Edge edge);

	/**
	 * Adds an outgoing edge to the node
	 * 
	 * @param edge
	 */
	public void addOutgoingEdgeAt(Edge edge, int index);

	/**
	 * adds a child as the new rightmost child of the node
	 * 
	 * @param child
	 * @return
	 */
	public Edge addChild(Node child, Expression expression);

	/**
	 * adds a child as the new rightmost child of the node
	 * 
	 * @param child
	 * @return
	 */
	public Edge addChild(Node child);

	/**
	 * adds a child as a new child of the node at the given index. if the given
	 * index is 0, then the new child is the leftmost child. If the given index
	 * is getOutgoingEdges().size() then the new child becomes the rightmost
	 * child.
	 * 
	 * @param child
	 * @return
	 */
	public Edge addChildAt(Node child, Expression expression, int index);

	/**
	 * adds a child as a new child of the node at the given index. if the given
	 * index is 0, then the new child is the leftmost child. If the given index
	 * is getOutgoingEdges().size() then the new child becomes the rightmost
	 * child.
	 * 
	 * @param child
	 * @return
	 */
	public Edge addChildAt(Node child, int index);
	
	/**
	 * removes the child at the given index and adds a new one in its place.
	 *  
	 * @param child
	 * @return
	 */
	public Edge swapChildAt(Node child, Expression expression, int index);
	
	/**
	 * removes the child at the given index and adds a new one in its place.
	 *  
	 * @param child
	 * @return
	 */
	public Edge swapChildAt(Node child, int index);

	/**
	 * returns an ordered list of nodes that are the children of this node in
	 * the tree
	 * 
	 * changes to the collection are NOT reflected in the tree
	 * 
	 * @return
	 */
	public List<Node> getChildren();

	/**
	 * returns an iterator over the children of a block. This iterator cannot be
	 * used to remove children from the block.
	 * 
	 * @return
	 */
	public Iterator<Node> iterator();

	/**
	 * returns the number of children.
	 * 
	 * @return
	 */
	public int numChildren();

	/**
	 * returns true if and only if the ordering of the children as reflected by
	 * the various methods and iterators matter.
	 * 
	 * @return
	 */
	public boolean orderingOfChildernMatters();

	/**
	 * returns true if and only if the expressions on outgoing edges matter.
	 * 
	 * @return
	 */
	public boolean expressionsOfOutgoingEdgesMatter();
	
	/**
	 * returns true if and only this block is changeable (note it is false by definition for events)
	 * 
	 * @return
	 */
	public boolean isChangeable();
	
	/**
	 * sets this block to be (non-)changeable
	 * 
	 * @param changeable
	 */
	public void setChangeable(boolean changeable);

	/**
	 * Denotes an Xor block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface Xor extends Block {

	}

	/**
	 * Denotes an Or block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface Or extends Block {

	}

	/**
	 * Denotes an And block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface And extends Block {

	}

	/**
	 * Denotes a sequence block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface Seq extends Block {

	}

	/**
	 * Denotes a Xor Loop block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface XorLoop extends Block {

	}

	/**
	 * Denotes an deferred choice block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface Def extends Block {

	}

	/**
	 * Denotes an deferred choice loop block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface DefLoop extends Block {

	}

	/**
	 * Denotes an PlaceHolder block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface PlaceHolder extends Block {

	}

}