package org.processmining.processtree.configuration.data.impl;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.processmining.framework.util.Pair;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Variable;
import org.processmining.processtree.configuration.ConfigurationAuxiliary;
import org.processmining.processtree.configuration.data.DataConfiguration;
import org.processmining.processtree.configuration.data.DataConfigurationIterator;
import org.processmining.processtree.impl.ProcessTreeImpl;

public class DataConfigurationIteratorImpl implements DataConfigurationIterator{

	private ProcessTree tree;
	public BitSet currentConfiguration;
	private List<UUID> readNodes;
	private List<UUID> writeNodes;
	private List<UUID> edges;
	public Map<UUID, List<UUID>> remRead;
	public Map<UUID, List<UUID>> remWrite;
	public Map<UUID, List<UUID>> remExpr;
	public BitSet lastReturnedConfiguration = null;
	private int bitLength;
	private BitSet minConfiguration = null;
	private BitSet maxConfiguration = null;
	private boolean doubleOption;

	public DataConfigurationIteratorImpl(ProcessTree tree) throws Exception{
		this(tree, null, null);
	}
	
	public DataConfigurationIteratorImpl(ProcessTree tree, Object minConfiguration, Object maxConfiguration){
		this.tree = new ProcessTreeImpl(tree);
		this.minConfiguration = (BitSet) minConfiguration;
		this.maxConfiguration = (BitSet) maxConfiguration;
		fillRemoved();		
		this.bitLength = getBitLength();
		sort();
		currentConfiguration = new BitSet(bitLength + 2); // last 2 bits is for denoting termination, and fixing the size
		currentConfiguration.clear();
		currentConfiguration.set(bitLength + 1, true);
		if(this.minConfiguration != null){
			currentConfiguration = (BitSet) this.minConfiguration.clone();
		}
	}
	
	private void fillRemoved(){
		remRead = new HashMap<UUID, List<UUID>>();
		remWrite = new HashMap<UUID, List<UUID>>();
		remExpr = new HashMap<UUID, List<UUID>>();
		readNodes = new ArrayList<UUID>();
		writeNodes = new ArrayList<UUID>();
		edges = new ArrayList<UUID>();
		for(Node n: tree.getNodes()){
			if(!n.getRemovableReadVariables().isEmpty()){
				// lets make the ArrayList
				List<UUID> reads = new ArrayList<UUID>();
				for(Variable var: n.getRemovableReadVariables()){
					reads.add(var.getID());
				}
				remRead.put(n.getID(), reads);
				readNodes.add(n.getID());
			}
			if(!n.getRemovableWrittenVariables().isEmpty()){
				// lets make the ArrayList
				List<UUID> writes = new ArrayList<UUID>();
				for(Variable var: n.getRemovableReadVariables()){
					writes.add(var.getID());
				}
				remWrite.put(n.getID(), writes);
				writeNodes.add(n.getID());
			}
		}
		for(Edge e: tree.getEdges()){
			if(!e.getRemovableExpressions().isEmpty()){
				// lets make the ArrayList
				List<UUID> exprs = new ArrayList<UUID>();
				for(Expression expr: e.getRemovableExpressions()){
					exprs.add(expr.getID());
				}
				remExpr.put(e.getID(), exprs);
				edges.add(e.getID());
			}
		}
	}
	
	private int getBitLength(){
		int bitLength = 0;
		for(UUID id: remRead.keySet()){
			bitLength += remRead.get(id).size();
		}
		for(UUID id: remWrite.keySet()){
			bitLength += remWrite.get(id).size();
		}
		for(UUID id: remExpr.keySet()){
			bitLength += remExpr.get(id).size();
		}
		return bitLength;
	}
	
	private void sort(){
		Collections.sort(readNodes);
		for(UUID id: readNodes){
			Collections.sort(remRead.get(id));
		}
		Collections.sort(writeNodes);
		for(UUID id: writeNodes){
			Collections.sort(remWrite.get(id));
		}
		Collections.sort(edges);
		for(UUID id: edges){
			Collections.sort(remExpr.get(id));
		}
	}
	
	public DataConfiguration nextConfiguration() throws Exception{
		DataConfiguration conf = new DataConfigurationImpl();
		if(lastReturnedConfiguration == null){
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		if(currentConfiguration.equals(lastReturnedConfiguration) && !currentConfiguration.get(currentConfiguration.length() - 2)){
			ConfigurationAuxiliary.nextBitSequece(currentConfiguration);
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		while((!isConsistentComplete(conf, tree) && !currentConfiguration.get(currentConfiguration.length() - 2) && ConfigurationAuxiliary.smaller(currentConfiguration, maxConfiguration))){
			ConfigurationAuxiliary.nextBitSequece(currentConfiguration);
			conf = new DataConfigurationImpl();
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		if(isConsistentComplete(conf, tree) && !currentConfiguration.get(currentConfiguration.length() - 2) && !(currentConfiguration.equals(lastReturnedConfiguration)) && ConfigurationAuxiliary.smaller(currentConfiguration, maxConfiguration)){
			lastReturnedConfiguration = (BitSet) currentConfiguration.clone();
			return conf;
		}
		else{
			return null;
		}
	}
	
	public boolean hasNext() throws Exception{
		BitSet currentC = (BitSet) currentConfiguration.clone();
		BitSet lastRetConf = null;
		if(lastReturnedConfiguration != null){
			lastRetConf = (BitSet) lastReturnedConfiguration.clone();
		}
		if(nextConfiguration() != null){
			currentConfiguration = currentC;
			lastReturnedConfiguration = lastRetConf;
			return true;
		}
		else{
			return false;
		}
	}
	
	public DataConfiguration getNext() throws Exception {
		if(hasNext()){
			return nextConfiguration();
		}
		throw new NullPointerException();
	}
	
	public void reset(){
		currentConfiguration = new BitSet(bitLength + 2); // last 2 bits is for denoting termination, and fixing the size
		currentConfiguration.clear();
		currentConfiguration.set(bitLength + 1, true);
		if(this.minConfiguration != null){
			currentConfiguration = (BitSet) this.minConfiguration.clone();
		}
		lastReturnedConfiguration = null;
	}
	
	private void fillConfiguration(DataConfiguration conf){
		// TODO: this can be done more efficient
		// note the -2 due to the control bits
		doubleOption = false;
		for(int i = 0; i < currentConfiguration.length() - 2; i++){
			// we need to have dynamic reduction on the length of the lists, i.e., we need to lookup to which nodes and variable this belongs
			int lowerboundOfInterval = 0;
			for(UUID id: edges){
				if(lowerboundOfInterval <= i && i < lowerboundOfInterval + remExpr.get(id).size()){
					if(currentConfiguration.get(i)){
						if(conf.getSelectedExpression(id) != null){
							doubleOption = true;
						}
						conf.setSelectedExpression(id, remRead.get(id).get(i - lowerboundOfInterval));
					}
				}
				else{
					lowerboundOfInterval += remRead.get(id).size();
				}
			}			
			for(UUID id: readNodes){
				if(lowerboundOfInterval <= i && i < lowerboundOfInterval + remRead.get(id).size()){
					if(currentConfiguration.get(i)){
						conf.addRemovedRead(id, remRead.get(id).get(i - lowerboundOfInterval));
					}
				}
				else{
					lowerboundOfInterval += remRead.get(id).size();
				}
			}
			for(UUID id: writeNodes){
				if(lowerboundOfInterval <= i && i < lowerboundOfInterval + remWrite.get(id).size()){
					if(currentConfiguration.get(i)){
						conf.addRemovedWritten(id, remWrite.get(id).get(i - lowerboundOfInterval));
					}
				}
				else{
					lowerboundOfInterval += remWrite.get(id).size();
				}
			}
		}
	}
	
	protected static void makeComplete(DataConfiguration conf, ProcessTree tree){
		for(Node n: tree.getNodes()){
			for(Variable var: n.getRemovableReadVariables()){
				if(!conf.getRemovedRead(n.getID()).contains(var.getID())){
					conf.addNonRemovedRead(n.getID(), var.getID());
				}
			}
			for(Variable var: n.getRemovableWrittenVariables()){
				if(!conf.getRemovedWritten(n.getID()).contains(var.getID())){
					conf.addNonRemovedWritten(n.getID(), var.getID());
				}
			}
		}
		for(Edge e: tree.getEdges()){
			for(Expression expr: e.getRemovableExpressions()){
				if(!conf.getSelectedExpression(e.getID()).equals(expr.getID())){
					conf.addNonSelectedExpression(e.getID(), expr.getID());
				}
			}
		}
	}
	
	private boolean isConsistentComplete(DataConfiguration conf, ProcessTree tree){
		for(UUID id: readNodes){
			if(ConfigurationAuxiliary.overlap(conf.getRemovedRead(id), conf.getNonRemovedRead(id))){
				return false;
			}
		}
		for(UUID id: writeNodes){
			if(ConfigurationAuxiliary.overlap(conf.getRemovedWritten(id), conf.getNonRemovedWritten(id))){
				return false;
			}
		}
		for(UUID id: edges){
			if(conf.getSelectedExpression(id) == null){
				return false;
			}
			if(ConfigurationAuxiliary.overlap(conf.getSelectedExpression(id), conf.getNonSelectedExpression(id))){
				return false;
			}
		}
		for(UUID id: edges){
			if(tree.getEdge(id).getRemovableExpressions().size() == conf.getNonSelectedExpression(id).size()){
				return false;
			}
		}
		if(doubleOption){
			return false;
		}
		return true;
	}

	@SuppressWarnings("unused")
	public List<Pair<Object, Object>> getSubdivisions(int divisions) throws Exception {
		// we have some bit fun here :-D
		List<Pair<Object, Object>> subdivisions = new ArrayList<Pair<Object,Object>>();
		BitSet min;
		BitSet max = new BitSet();
		// easy, we get this one for free :D
		min = (BitSet) currentConfiguration.clone();
		throw new Exception("Not implemented\n");
		//return null;
	}

}
