package org.processmining.processtree.conversion.yawl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.processtree.Edge;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.configuration.Configuration;
import org.processmining.processtree.configuration.controlflow.ControlFlowConfiguration;
import org.yawlfoundation.yawl.editor.elements.model.EngineIdentifier;
import org.yawlfoundation.yawl.editor.specification.SpecificationModel;

@SuppressWarnings("deprecation")
@Plugin(
		name = "Convert Process Tree and Configruation to a YAWL model, Domain Configuration, and a CMapping", 
		parameterLabels = { "Process Tree", "Configuration" }, 
		returnLabels = { "YAWL Model", "Domain Configuration", "CMapping" }, 
		returnTypes = { SpecificationModel.class, DomainConfiguration.class, CMapping.class }, 
		userAccessible = true)
public class Configuration2YAWL {
	HashMap<Edge, String> factsHidden = new HashMap<Edge, String>(); // the ids of the facts for the edges configured to be hidden
	HashMap<Edge, String> factsBlocked = new HashMap<Edge, String>(); // the ids of the facts for the edges configured to be blocked
	HashMap<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds = new HashMap<Edge, Pair<EngineIdentifier,EngineIdentifier>>(); // the ids of from and to for edges (from and to are the YAWL identifiers)
	int facId = 1;
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "D.M.M. Schunselaar", email = "d.m.m.schunselaar@tue.nl", uiLabel = "Convert ProcessTree YAWL (Model, Configuration, CMapping)")
	@PluginVariant(variantLabel = "Convert Process Tree to YAWL elements", requiredParameterLabels = { 0, 1 })
	public Object convertConfiguration2YAWL(PluginContext context, ProcessTree tree, Configuration conf) throws InstantiationException, IllegalAccessException{
		return convertConfiguration2YAWL(tree, conf, "ConfigurableYAWLModel.yawl");
	}
	
	
	public Object[] convertConfiguration2YAWL(ProcessTree tree, Configuration conf, String fileName) throws InstantiationException, IllegalAccessException{
		// so we need to generate 2 files simultaneous, and keep all the ids consistent, the fun... -.-
		ProcessTree2Yawl pt2y = new ProcessTree2Yawl();
		SpecificationModel specification = pt2y.convertConfigurableModel(tree, mappingEdgeIds);
		// so we have the specification and the mapping from edges to EngineIds needed for the other parts
		DomainConfiguration domainConfig = new DomainConfiguration(fileName);
		CMapping cMap = new CMapping(fileName);
		
		// now lets create all the different facts
		for(Edge e: tree.getEdges()){
			domainConfig.append(createFact(e, conf.getControlFlowConfiguration()));
		}
		domainConfig.finalise();
		
		// we are done with domain configuration, now make the mapping
		List<String> splits = new ArrayList<String>();
		List<String> joins = new ArrayList<String>();
		for(Edge e: tree.getEdges()){
			splits.add(createSplit(e));
			joins.add(createJoin(e));
		}
		cMap.appendSplits(splits);
		cMap.appendJoins(joins);
		cMap.finalise();
		
		Object[] ret = new Object[3];
		ret[0] = specification;
		ret[1] = domainConfig;
		ret[2] = cMap;
		return ret;
	}

	private String createFact(Edge e, ControlFlowConfiguration conf) {
		// so we have the edge, lets make the different facts
		String ret = "";
		if(e.isHideable() && (conf.isHidden(e) || conf.isNonHidden(e))){
			String factId = "f" + facId++;
			factsHidden.put(e, factId);
			ret += "<Fact id=\"" + factId + "\">\n";
			ret += "\t<description>Edge " + e.getName() + " is hidden</description>\n";
			ret += "\t<value>" + conf.isHidden(e) + "</value>\n";
			// TODO: I have no idea what the next line means
			ret += "\t<deviates>true</deviates>\n";
			ret += "</Fact>\n";
		}
		if(e.isBlockable() && (conf.isBlocked(e) || conf.isNonBlocked(e))){
			String factId = "f" + facId++;
			factsBlocked.put(e, factId);
			ret += "<Fact id=\"" + factId + "\">\n";
			ret += "\t<description>Edge " + e.getName() + " is blocked</description>\n";
			ret += "\t<value>" + conf.isBlocked(e) + "</value>\n";
			// TODO: I have no idea what the next line means
			ret += "\t<deviates>true</deviates>\n";
			ret += "</Fact>\n";
		}
		return ret;
	}
	

	private String createSplit(Edge e) {
		String ret = "";
		// first check if this edge is configured
		if(factsBlocked.containsKey(e)){
			// now we need the ids
			ret += "\t\t\t<port sourceId=\"" + mappingEdgeIds.get(e).getFirst().toString() + "\" targetId=\"" + mappingEdgeIds.get(e).getSecond().toString() + "\">\n";
			ret += "\t\t\t\t<value condition=\"" + factsBlocked.get(e) + " = true\" type=\"blocked\"/>\n";
			ret += "\t\t\t\t<value condition=\"" + factsBlocked.get(e) + " = false\" type=\"activated\"/>\n";
			ret += "\t\t\t</port>\n";
		}
		return ret;
	}
	
	private String createJoin(Edge e) {
		String ret = "";
		// first check if this edge is configured
		if(factsBlocked.containsKey(e) || factsHidden.containsKey(e)){
			// now we need the ids
			ret += "\t\t\t<port sourceId=\"" + mappingEdgeIds.get(e).getFirst().toString() + "\" targetId=\"" + mappingEdgeIds.get(e).getSecond().toString() + "\">\n";
			// now we have to encode the 3-state logic
			if(factsBlocked.containsKey(e)){
				ret += "\t\t\t\t<value condition=\"" + factsBlocked.get(e) + " = true\" type=\"blocked\"/>\n";
			}
			if(factsHidden.containsKey(e)){
				ret += "\t\t\t\t<value condition=\"" + factsHidden.get(e) + " = true\" type=\"hidden\"/>\n";
			}
			// now we have the fun, which are present
			String subString = "";
			if(factsBlocked.containsKey(e)){
				subString += factsBlocked.get(e) + " = false";
			}
			if(factsBlocked.containsKey(e) && factsHidden.containsKey(e)){
				subString += " . ";
			}
			if(factsHidden.containsKey(e)){
				subString += factsHidden.get(e) + " = false";
			}
			ret += "\t\t\t\t<value condition=\"" + subString + "\" type=\"activated\"/>\n";
			ret += "\t\t\t</port>\n";
		}
		return ret;
	}
	
}
