package org.processmining.processtree.ptml;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Event;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.Originator;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Task;
import org.processmining.processtree.Variable;
import org.xmlpull.v1.XmlPullParser;

public class PtmlProcessTree extends PtmlProcessTreeElement {

	/**
	 * PTML tag.
	 */
	public final static String TAG = "processTree";

	private UUID root;
	
	private List<PtmlNode> nodeList;
	private List<PtmlIsParent> isParentList;
	
	private List<PtmlOriginator> originatorList;
	private List<PtmlIsOriginator> isOriginatorList;

	private List<PtmlVariable> variableList;
	private List<PtmlIsUsed> isUsedList;
	private List<PtmlIsRead> isReadList;
	private List<PtmlIsWritten> isWrittenList;
	private List<PtmlIsExpression> isExpressionList;
	
	private List<PtmlExpression> expressionList;
		
	public PtmlProcessTree(UUID id) {
		super(TAG, id);
		name = null;
		root = null;
		nodeList = new ArrayList<PtmlNode>();
		isParentList = new ArrayList<PtmlIsParent>();
		originatorList = new ArrayList<PtmlOriginator>();
		isOriginatorList = new ArrayList<PtmlIsOriginator>();
		variableList = new ArrayList<PtmlVariable>();
		isUsedList = new ArrayList<PtmlIsUsed>();
		isReadList = new ArrayList<PtmlIsRead>();
		isWrittenList = new ArrayList<PtmlIsWritten>();
		isExpressionList = new ArrayList<PtmlIsExpression>();
		expressionList = new ArrayList<PtmlExpression>();
	}

	protected void importAttributes(XmlPullParser xpp, Ptml ptml) {
		super.importAttributes(xpp, ptml);
		importRoot(xpp, ptml);
	}

	private void importRoot(XmlPullParser xpp, Ptml ptml) {
		String value = xpp.getAttributeValue(null, "root");
		if (value != null) {
			root = UUID.fromString(value);
		}
		ptml.log(root != null ? value : "<no root>");
	}

	protected String exportAttributes(Ptml ptml) {
		return super.exportAttributes(ptml) + exportRoot(ptml);
	}

	private String exportRoot(Ptml ptml) {
		if (root != null) {
			return exportAttribute("root", root.toString(), ptml);
		}
		return "";
	}

	protected boolean importElements(PluginContext context, XmlPullParser xpp, Ptml ptml) {
		if (super.importElements(context, xpp, ptml)) {
			return true;
		} else {
			PtmlNode node = null;
			if (xpp.getName().equals(PtmlAnd.TAG)) {
				node = new PtmlAnd(null);
			} else if (xpp.getName().equals(PtmlDef.TAG)) {
				node = new PtmlDef(null);
			} else if (xpp.getName().equals(PtmlDefLoop.TAG)) {
				node = new PtmlDefLoop(null);
			} else if (xpp.getName().equals(PtmlOr.TAG)) {
				node = new PtmlOr(null);
			} else if (xpp.getName().equals(PtmlPlaceHolder.TAG)) {
				node = new PtmlPlaceHolder(null);
			} else if (xpp.getName().equals(PtmlSeq.TAG)) {
				node = new PtmlSeq(null);
			} else if (xpp.getName().equals(PtmlXor.TAG)) {
				node = new PtmlXor(null);
			} else if (xpp.getName().equals(PtmlXorLoop.TAG)) {
				node = new PtmlXorLoop(null);
			} else if (xpp.getName().equals(PtmlAutomatic.TAG)) {
				node = new PtmlAutomatic(null);
			} else if (xpp.getName().equals(PtmlManual.TAG)) {
				node = new PtmlManual(null);
			} else if (xpp.getName().equals(PtmlTimeOut.TAG)) {
				node = new PtmlTimeOut(null);
			} else if (xpp.getName().equals(PtmlMessage.TAG)) {
				node = new PtmlMessage(null);
			}
			if (node != null) {
				nodeList.add(node);
				node.importElement(context, xpp, ptml);
				return true;
			}
			if (xpp.getName().equals(PtmlIsParent.TAG)) {
				PtmlIsParent isParent = new PtmlIsParent(null);
				isParentList.add(isParent);
				isParent.importElement(context, xpp, ptml);
				return true;
			}
			PtmlOriginator originator = null;
			if (xpp.getName().equals(PtmlGroup.TAG)) {
				originator = new PtmlGroup(null);
			} else if (xpp.getName().equals(PtmlResource.TAG)) {
				originator = new PtmlResource(null);
			} else if (xpp.getName().equals(PtmlRole.TAG)) {
				originator = new PtmlRole(null);
			}
			if (originator != null) {
				originatorList.add(originator);
				originator.importElement(context, xpp, ptml);
				return true;
			}
			if (xpp.getName().equals(PtmlIsOriginator.TAG)) {
				PtmlIsOriginator isOriginator = new PtmlIsOriginator(null);
				isOriginatorList.add(isOriginator);
				isOriginator.importElement(context, xpp, ptml);
				return true;
			} else if (xpp.getName().equals(PtmlVariable.TAG)) {
				PtmlVariable variable = new PtmlVariable(null);
				variableList.add(variable);
				variable.importElement(context, xpp, ptml);
				return true;
			} else if (xpp.getName().equals(PtmlIsUsed.TAG)) {
				PtmlIsUsed isUsed = new PtmlIsUsed(null);
				isUsedList.add(isUsed);
				isUsed.importElement(context, xpp, ptml);
				return true;
			} else if (xpp.getName().equals(PtmlIsRead.TAG)) {
				PtmlIsRead isRead = new PtmlIsRead(null);
				isReadList.add(isRead);
				isRead.importElement(context, xpp, ptml);
				return true;
			} else if (xpp.getName().equals(PtmlIsWritten.TAG)) {
				PtmlIsWritten isWritten = new PtmlIsWritten(null);
				isWrittenList.add(isWritten);
				isWritten.importElement(context, xpp, ptml);
				return true;
			}
			if(xpp.getName().equals(PtmlExpression.TAG)){
				PtmlExpression expression = new PtmlExpression(null);
				expressionList.add(expression);
				expression.importElement(context, xpp, ptml);
				return true;
			}
			if (xpp.getName().equals(PtmlIsExpression.TAG)){
				PtmlIsExpression isExpression = new PtmlIsExpression(null);
				isExpressionList.add(isExpression);
				isExpression.importElement(context, xpp, ptml);
				return true;
			}
			
		}
		return false;
	}

	protected String exportElements(Ptml ptml) {
		String s = super.exportElements(ptml);
		//System.out.println("Before exporting: " + nodeList.size() + "," + isParentList.size() + "," + originatorList.size() + "," + isOriginatorList.size());
		for (PtmlNode node : nodeList) {
			s += node.exportElement(ptml);
		}
		for (PtmlEdge edge : isParentList) {
			s += edge.exportElement(ptml);
		}
		for (PtmlOriginator originator : originatorList) {
			s += originator.exportElement(ptml);
		}
		for (PtmlIsOriginator isOriginator : isOriginatorList) {
			s += isOriginator.exportElement(ptml);
		}
		for(PtmlExpression expression: expressionList){
			s += expression.exportElement(ptml);
		}
		for (PtmlVariable variable : variableList) {
			s += variable.exportElement(ptml);
		}
		for (PtmlIsUsed isUsed : isUsedList) {
			s += isUsed.exportElement(ptml);
		}
		for (PtmlIsRead isRead : isReadList) {
			s += isRead.exportElement(ptml);
		}
		for (PtmlIsWritten isWritten : isWrittenList) {
			s += isWritten.exportElement(ptml);
		}
		for (PtmlIsExpression isExpression : isExpressionList) {
			s += isExpression.exportElement(ptml);
		}
		
		return s;
	}

	public void unmarshall(ProcessTree tree, Map<UUID, Node> nodeMap, Map<UUID, Edge> edgeMap, Map<UUID, Originator> originatorMap, Map<UUID, Variable> variableMap, Map<UUID, Expression> expressionMap) {
		for (PtmlNode node : nodeList) {
			node.unmarshall(tree, nodeMap);
		}
		if (nodeMap.containsKey(root)) {
			tree.setRoot(nodeMap.get(root));
		}
		for(PtmlExpression expr: expressionList){
			expr.unmarshall(tree, expressionMap);
		}
		for (PtmlIsParent isParent : isParentList) {
			isParent.unmarshall(tree, nodeMap, edgeMap, expressionMap);
		}
		for (PtmlOriginator originator : originatorList) {
			originator.unmarshall(tree, originatorMap);
		}
		for (PtmlIsOriginator isOriginator : isOriginatorList) {
			isOriginator.unmarshall(tree, nodeMap, originatorMap);
		}
		for (PtmlVariable variable : variableList) {
			variable.unmarshall(tree, variableMap);
		}
		for (PtmlIsUsed isUsed : isUsedList) {
			//isUsed.unmarshall(tree, edgeMap, variableMap);
			isUsed.unmarshall(tree, expressionMap, variableMap);
		}
		for (PtmlIsRead isRead : isReadList) {
			isRead.unmarshall(tree, nodeMap, variableMap);
		}
		for (PtmlIsWritten isWritten : isWrittenList) {
			isWritten.unmarshall(tree, nodeMap, variableMap);
		}
		for (PtmlIsExpression isExpression : isExpressionList) {
			isExpression.unmarshall(tree, edgeMap, expressionMap);
		}
		if(propertyIndependent != null){
			propertyIndependent.unmarshall(tree);
		}
		if(propertyDependent != null){
			propertyDependent.unmarshall(tree);
		}
//		System.out.println("After unmarshalling: " + nodeList.size() + "," + isParentList.size() + "," + originatorList.size() + "," + isOriginatorList.size());
	}

	public void marshall(ProcessTree tree) {
		this.id = tree.getID();
		if(tree.getName() != null){
			this.name = tree.getName();
		}
		else{
			this.name = "NoName";
		}
		for (Node node : tree.getNodes()) {
			PtmlNode n = null;
			if (node instanceof Block.And) {
				n = new PtmlAnd(node.getID());
			} else if (node instanceof Block.Def) {
				n = new PtmlDef(node.getID());
			} else if (node instanceof Block.DefLoop) {
				n = new PtmlDefLoop(node.getID());
			} else if (node instanceof Block.Or) {
				n = new PtmlOr(node.getID());
			} else if (node instanceof Block.PlaceHolder) {
				n = new PtmlPlaceHolder(node.getID());
			} else if (node instanceof Block.Seq) {
				n = new PtmlSeq(node.getID());
			} else if (node instanceof Block.Xor) {
				n = new PtmlXor(node.getID());
			} else if (node instanceof Block.XorLoop) {
				n = new PtmlXorLoop(node.getID());
			} else if (node instanceof Task.Automatic) {
				n = new PtmlAutomatic(node.getID());
			} else if (node instanceof Task.Manual) {
				n = new PtmlManual(node.getID());
			} else if (node instanceof Event.TimeOut) {
				n = new PtmlTimeOut(node.getID());
			} else if (node instanceof Event.Message) {
				n = new PtmlMessage(node.getID());
			}
			if (n != null) {
				if(!(n instanceof PtmlBlock)){
					n.marshall(tree, node);
				}
				else{
					((PtmlBlock)n).marshall(tree, (Block)node);
				}
				nodeList.add(n);
			}
		}
		root = tree.getRoot().getID();
		Map<Edge, UUID> edgeMap = new HashMap<Edge, UUID>();
		for (Node node : tree.getNodes()) {
			if (node instanceof Block) {
				Block block = (Block) node;
				for (Edge edge : block.getOutgoingEdges()) {
					PtmlIsParent isParent = new PtmlIsParent(edge.getID());
					isParent.marshall(tree, edge, edgeMap);
					isParentList.add(isParent);
				}
			}
		}
		Map<Originator, UUID> originatorMap = new HashMap<Originator, UUID>();
		for (Originator originator : tree.getOriginators()) {
			PtmlOriginator o = null;
			if (originator instanceof Originator.Group) {
				o = new PtmlGroup(null);
			} else if (originator instanceof Originator.Resource) {
				o = new PtmlResource(null);
			} else if (originator instanceof Originator.Role) {
				o = new PtmlRole(null);
			}
			if (o != null) {
				o.marshall(tree, originator, originatorMap);
				originatorList.add(o);
			}
		}
		for (Node node : tree.getNodes()) {
			if (node instanceof Task.Manual) {
				for (Originator originator : ((Task.Manual) node).getOriginators()) {
					PtmlIsOriginator isOriginator = new PtmlIsOriginator(null);
					isOriginator.marshall(tree, (Task.Manual) node, originator, originatorMap); 
					isOriginatorList.add(isOriginator);
				}
			}
		}
		Map<Variable, UUID> variableMap = new HashMap<Variable, UUID>();
		for (Variable variable : tree.getVariables()) {
			PtmlVariable v = new PtmlVariable(null);
			v.marshall(tree, variableMap, variable);
			variableList.add(v);
		}
		
		Map<Expression, UUID> expressionMap = new HashMap<Expression, UUID>();
		for(Expression expr: tree.getExpressions()){
			PtmlExpression ex = new PtmlExpression(null);
			ex.marshall(tree, expressionMap, expr);
			expressionList.add(ex);
		}
		
		for(Expression expr: tree.getExpressions()){
			for(Variable variable: expr.getVariables()){
				PtmlIsUsed isUsed = new PtmlIsUsed(null);
				isUsed.marshall(tree, expr, expressionMap, variable, variableMap);
				isUsedList.add(isUsed);
			}
		}
		for (Node node : tree.getNodes()) {
			for (Variable variable : node.getReadVariables()) {
				PtmlIsRead isRead = new PtmlIsRead(null);
				isRead.marshall(tree, node, variable, variableMap);
				isReadList.add(isRead);
			}
			for (Variable variable : node.getWrittenVariables()) {
				PtmlIsWritten isWritten = new PtmlIsWritten(null);
				isWritten.marshall(tree, node, variable, variableMap);
				isWrittenList.add(isWritten);
			}
		}
		for(Edge e : tree.getEdges()) {
			for(Expression expr : e.getExpressions()){
				PtmlIsExpression isExpression = new PtmlIsExpression(null);
				isExpression.marshall(tree, e, expr, expressionMap);
				isExpressionList.add(isExpression);
			}
		}
		if(tree.getIndependentProperties() != null && !tree.getIndependentProperties().isEmpty()){
			propertyIndependent = new PtmlPropertyIndependent();
			propertyIndependent.marshall(tree.getIndependentProperties());
		}
		if(tree.getDependentProperties() != null && !tree.getDependentProperties().isEmpty()){
			propertyDependent = new PtmlPropertyDependent();
			propertyDependent.marshall(tree.getDependentProperties());
		}
//		System.out.println("After marshalling: " + nodeList.size() + "," + isParentList.size() + "," + originatorList.size() + "," + isOriginatorList.size());
	}
	
	private Set<UUID> getSourceIds(List<PtmlIsParent> list) {
		Set<UUID> set = new HashSet<UUID>();
		for (PtmlIsParent element : list) {
			set.add(element.sourceId);
		}
		return set;
	}
	
	private Set<UUID> getTargetIds(List<PtmlIsParent> list) {
		Set<UUID> set = new HashSet<UUID>();
		for (PtmlIsParent element : list) {
			set.add(element.targetId);
		}
		return set;
	}
	
	private List<PtmlIsParent> getCyclicEdges() {
		List<PtmlIsParent> list = new ArrayList<PtmlIsParent>(isParentList);
		int oldSize = Integer.MAX_VALUE;
		while (!list.isEmpty() && (list.size() < oldSize)) {
			Set<UUID> sources = getSourceIds(list);
			Set<UUID> targets = getTargetIds(list);
			Set<UUID> roots = new HashSet<UUID>(sources);
			roots.removeAll(targets);
			List<PtmlIsParent> newList = new ArrayList<PtmlIsParent>();
			for (PtmlIsParent element : list) {
				if (!roots.contains(element.sourceId)) {
					newList.add(element);
				}
			}
			oldSize = list.size();
			list = newList;
		}
		return list;
	}
	
	public void checkAcyclicity(Ptml ptml) {
		int i = 0;
		for (PtmlIsParent edge : getCyclicEdges()) {
			ptml.log(TAG, i++, "Cyclic edge: " + edge.sourceId + " -> " + edge.targetId);
		}
	}
}
