package org.processmining.projectedrecallandprecision.helperclasses;

import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.Iterator;

import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.EfficientPetrinetSemantics;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.impl.EfficientPetrinetSemanticsImpl;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAFactory;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Minimise;

import gnu.trove.iterator.TShortObjectIterator;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.TShortObjectMap;
import gnu.trove.map.custom_hash.TObjectIntCustomHashMap;
import gnu.trove.map.hash.TShortObjectHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.strategy.HashingStrategy;

public class AcceptingPetriNet2automaton {

	public static final HashingStrategy<byte[]> hashingStrategy = new HashingStrategy<byte[]>() {
		private static final long serialVersionUID = 1L;

		public int computeHashCode(byte[] object) {
			return Arrays.hashCode(object);
		}

		public boolean equals(byte[] o1, byte[] o2) {
			return Arrays.equals(o1, o2);
		}
	};

	/**
	 * The ReachabilityGraph class of ProM gives errors.
	 * 
	 * Idea: construct an automaton directly.
	 * 
	 * @param net
	 * @param canceller
	 * @throws AutomatonFailedException
	 */

	public static DFA convert(AcceptingPetriNet net, long maxStates, ProMCanceller canceller)
			throws AutomatonFailedException {
		DFA result = DFAFactory.createAutomaton();

		TObjectIntMap<byte[]> markingSet2state = new TObjectIntCustomHashMap<>(hashingStrategy, 10, 0.5f, -1); //set

		EfficientPetrinetSemantics semantics = new EfficientPetrinetSemanticsImpl(net.getNet(),
				net.getInitialMarking());

		//transform the final markings
		TCustomHashSet<byte[]> finalMarkings = new TCustomHashSet<>(hashingStrategy); //markings; not sets
		for (Marking finalMarking : net.getFinalMarkings()) {
			finalMarkings.add(semantics.convert(finalMarking));
		}

		ArrayDeque<byte[]> worklist = new ArrayDeque<>(); //deque of set

		byte[] initialMarking = semantics.convert(net.getInitialMarking());
		byte[] initialMarkingSet = exploreSilentTransitions(initialMarking, semantics, net);
		markingSet2state.put(initialMarkingSet, result.getInitialState());
		worklist.add(initialMarkingSet);

		//process all reachable markings
		long statesCounter = 0;
		while (!worklist.isEmpty()) {
			byte[] sourceMarkingSet = worklist.remove();
			int resultSourceState = markingSet2state.get(sourceMarkingSet);

			result.setAccepting(resultSourceState,
					AcceptingPetriNetMarkingSet.containsFinalMarking(sourceMarkingSet, finalMarkings));

			//gather the next states
			TShortObjectMap<byte[]> nextMarkingSets = new TShortObjectHashMap<>(10, 0.5f, (short) -1);

			for (Iterator<byte[]> it = AcceptingPetriNetMarkingSet.getMarkings(sourceMarkingSet); it.hasNext();) {
				byte[] marking = it.next();

				//				System.out.println("Consider new marking " + Arrays.toString(marking) + " . States seen: "
				//						+ statesCounter + ", states in queue: " + worklist.size());
				for (Transition transition : net.getNet().getTransitions()) {
					semantics.setState(marking);
					if (!transition.isInvisible() && semantics.isEnabled(transition)) {
						//execute the transition and record the new marking
						semantics.directExecuteExecutableTransition(transition);
						byte[] newMarking = semantics.getState();
						byte[] newMarkingSet = exploreSilentTransitions(newMarking, semantics, net);

						short activity = (short) transition.getLabel().charAt(0);
						//						System.out.println(" transition " + activity + " to marking " + Arrays.toString(newMarking)
						//								+ ", leading with silent steps to " + Arrays.toString(newMarkingSet));

						if (nextMarkingSets.containsKey(activity)) {
							//there was already an outgoing edge present this round
							nextMarkingSets.put(activity,
									AcceptingPetriNetMarkingSet.addAll(nextMarkingSets.get(activity), newMarkingSet));
						} else {
							nextMarkingSets.put(activity, newMarkingSet);
						}
						//						System.out.println("  activity " + activity + " now leading to " + Arrays.toString(nextMarkingSets.get(activity)));
					}

					if (canceller.isCancelled()) {
						return null;
					}
				}
			}

			//add states to result automaton
			for (TShortObjectIterator<byte[]> it = nextMarkingSets.iterator(); it.hasNext();) {
				it.advance();
				byte[] stateSetTarget = it.value();
				short activity = it.key();
				int resultTargetState = markingSet2state.get(stateSetTarget);
				if (resultTargetState == markingSet2state.getNoEntryValue()) {
					//this state pair did not exist yet
					resultTargetState = result.addEdge(resultSourceState, activity);
					markingSet2state.put(stateSetTarget, resultTargetState);

					worklist.add(stateSetTarget);
					statesCounter++;
				} else {
					//this state pair did not exist yet
					result.addEdge(resultSourceState, activity, resultTargetState);
				}
			}

			if (canceller.isCancelled()) {
				return null;
			}
		}

		if (statesCounter > maxStates) {
			throw new AutomatonFailedException("too many states in the reachability graph");
		}

		Automaton2Minimise.minimise(result, canceller);
		if (canceller.isCancelled()) {
			return null;
		}

		return result;
	}

	/**
	 * 
	 * @param marking
	 * @param semantics
	 * @param net
	 * @return A marking set of all markings reachable with silent transitions
	 *         from marking (includes marking itself).
	 */
	private static byte[] exploreSilentTransitions(byte[] marking, EfficientPetrinetSemantics semantics,
			AcceptingPetriNet net) {
		byte[] result = AcceptingPetriNetMarkingSet.create(marking);

		ArrayDeque<byte[]> worklist = new ArrayDeque<>();
		worklist.add(marking);

		while (!worklist.isEmpty()) {
			byte[] sourceMarking = worklist.remove();
			for (Transition transition : net.getNet().getTransitions()) {
				semantics.setState(sourceMarking);
				if (transition.isInvisible() && semantics.isEnabled(transition)) {
					semantics.directExecuteExecutableTransition(transition);
					byte[] targetMarking = semantics.getState();

					if (!AcceptingPetriNetMarkingSet.contains(result, targetMarking)) {
						result = AcceptingPetriNetMarkingSet.add(result, targetMarking);
						worklist.add(sourceMarking);
					}
				}
			}
		}

		return result;
	}

}
