package org.processmining.projectedrecallandprecision.helperclasses;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.commons.math3.util.CombinatoricsUtils;

/**
 * Walk over the binomial combinations of a list of elements. Thread safe, as
 * long as each thread provides a unique thread number.
 * 
 * @author sander
 *
 * @param <X>
 */
public class BinomialIteratorThreadSafe<X> {

	//input list
	final List<X> list;
	final Class<X> clazz;
	final int cardinality;

	final AtomicLong nextCombination;
	final long combinations;

	final List<Iterator<int[]>> threadIterators;
	final int[] threadCombinations;

	public BinomialIteratorThreadSafe(Class<X> clazz, List<X> list, int cardinality, int numberOfThreads) {
		this.list = list;
		this.clazz = clazz;
		this.cardinality = cardinality;
		nextCombination = new AtomicLong(-1);

		combinations = CombinatoricsUtils.binomialCoefficient(list.size(), cardinality);
		threadIterators = new ArrayList<>(numberOfThreads);
		threadCombinations = new int[numberOfThreads];
		for (int thread = 0; thread < numberOfThreads; thread++) {
			if (list.size() > cardinality) {
				threadIterators.add(CombinatoricsUtils.combinationsIterator(list.size(), cardinality));
			} else {
				//the combinatorics library gives an error if list.size == cardinality
				int[] x = new int[list.size()];
				for (int i = 0; i < list.size(); i++) {
					x[i] = i;
				}
				List<int[]> ar = new ArrayList<>(1);
				ar.add(x);
				threadIterators.add(ar.iterator());
			}
			threadCombinations[thread] = -1;
		}
	}

	/**
	 * 
	 * @param threadNr
	 * @return the number of the combination that is to be processed next.
	 *         Guaranteed to be thread safe. Returns null if all combinations
	 *         have been processed.
	 */
	public X[] next(int threadNr, boolean debug, int debugEvery) {
		long nowCombination = nextCombination.incrementAndGet();

		if (nowCombination >= combinations) {
			return null;
		}

		if (debug && nowCombination % debugEvery == 0) {
			System.out.println("start combination " + nowCombination + " of " + combinations);
		}

		Iterator<int[]> it = threadIterators.get(threadNr);
		int[] combination = null;
		while (threadCombinations[threadNr] < nowCombination) {
			threadCombinations[threadNr]++;
			combination = it.next();
		}

		//transform the int[] to X[]
		@SuppressWarnings("unchecked")
		X[] result = (X[]) Array.newInstance(clazz, cardinality);
		for (int i = 0; i < cardinality; i++) {
			result[i] = list.get(combination[i]);
		}

		return result;
	}
}
