package org.processmining.projectedrecallandprecision.helperclasses.automata;

import java.util.BitSet;

import gnu.trove.TCollections;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.iterator.TShortIterator;
import gnu.trove.list.TIntList;
import gnu.trove.list.TShortList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.list.array.TShortArrayList;

public class DFAImplEdgeList implements DFA {

	private int maxState;
	private final BitSet accepting;

	private final TIntArrayList sources;
	private final TIntArrayList targets;
	private final TShortArrayList activities;

	public DFAImplEdgeList() {
		maxState = 0;
		sources = new TIntArrayList();
		targets = new TIntArrayList();
		activities = new TShortArrayList();
		accepting = new BitSet();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#getInitialState()
	 */
	@Override
	public int getInitialState() {
		return 0;
	}

	/**
	 * 
	 * @return The number of the newly created state.
	 */
	private int addState() {
		maxState++;
		return maxState;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#setAccepting(int, boolean)
	 */
	@Override
	public void setAccepting(int state, boolean accepting) {
		this.accepting.set(state, accepting);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#isAccepting(int)
	 */
	@Override
	public boolean isAccepting(int state) {
		return accepting.get(state);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#getAcceptingStates()
	 */
	@Override
	public BitSet getAcceptingStates() {
		return (BitSet) accepting.clone();
	}

	/**
	 * Used to sort the edges.
	 * 
	 * @param source1
	 * @param target1
	 * @param activity1
	 * @param source2
	 * @param target2
	 * @param activity2
	 * @return
	 */
	private int compare(int source1, short activity1, int source2, short activity2) {
		if (source1 < source2) {
			return 1;
		} else if (source1 > source2) {
			return -1;
		} else if (activity1 < activity2) {
			return 1;
		} else if (activity1 > activity2) {
			return -1;
		} else {
			return 0;
		}
	}

	/**
	 * Similar to Arrays.binarySearch, but works on both sources and targets
	 * 
	 * @param source
	 * @param target
	 * @return
	 */
	private int binarySearch(int source, short activity) {
		int low = 0;
		int high = sources.size() - 1;
		int mid;
		int midVal;

		while (low <= high) {
			mid = (low + high) >>> 1;
			midVal = compare(source, activity, sources.get(mid), activities.get(mid));

			if (midVal < 0)
				low = mid + 1;
			else if (midVal > 0)
				high = mid - 1;
			else
				return mid; // key found
		}
		return -(low + 1); // key not found.
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#addEdge(int, short, int)
	 */
	@Override
	public void addEdge(int source, short activity, int target) {
		assert (source >= 0);
		assert (target >= 0);

		maxState = Math.max(maxState, target);

		//idea: keep the sources and targets sorted; first on source then on activity
		int from = binarySearch(source, activity);
		if (from >= 0) {
			//edge already present; replace target
			targets.set(from, target);
		} else {
			sources.insert(~from, source);
			targets.insert(~from, target);
			activities.insert(~from, activity);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#addEdge(int, short)
	 */
	@Override
	public int addEdge(int source, short activity) {
		assert (source >= 0);

		int from = binarySearch(source, activity);
		if (from >= 0) {
			//edge already present; return target
			return targets.get(from);
		} else {
			//edge not present; insert
			int target = addState();
			sources.insert(~from, source);
			targets.insert(~from, target);
			activities.insert(~from, activity);
		}

		EdgeIterableOutgoing iterator = new EdgeIterableOutgoing(source);
		while (iterator.hasNext()) {
			int target = iterator.nextTarget();
			if (iterator.getActivity() == activity) {
				return target;
			}
		}
		int target = addState();
		addEdge(source, activity, target);
		return target;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#replaceEdgesTo(int, int)
	 */
	@Override
	public void removeStateAndRedirectEdgesTo(int oldTarget, int newTarget) {
		assert (oldTarget > newTarget);
		for (int edge = 0; edge < targets.size(); edge++) {
			if (targets.get(edge) == oldTarget) {
				targets.set(edge, newTarget);
			}
		}

		/*
		 * Remove the old state, i.e. all its outgoing edges.
		 */
		removeOutgoingEdgesOf(oldTarget);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#removeState(int)
	 */
	@Override
	public void removeOutgoingEdgesOf(int state) {
		accepting.set(state, false);

		//from = the index of the first edge with oldTarget as source.
		int from = sources.binarySearch(state);
		if (from < 0) {
			return;
		}
		while (from >= 0 && sources.get(from) == state) {
			from--;
		}
		from++;

		//to = the index of the first edge -not- with oldTarget as source.
		int to = from + 1;
		while (to < sources.size() && sources.get(to) == state) {
			to++;
		}

		sources.remove(from, to - from);
		targets.remove(from, to - from);
		activities.remove(from, to - from);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#removeStatesExcept(java.util.BitSet)
	 */
	@Override
	public void removeStatesExcept(BitSet keepStates) {
		accepting.and(keepStates);
		TIntIterator itSources = sources.iterator();
		TIntIterator itTargets = targets.iterator();
		TShortIterator itActivities = activities.iterator();
		while (itSources.hasNext()) {
			int source = itSources.next();
			int target = itTargets.next();
			itActivities.next();

			if (!keepStates.get(source) || !keepStates.get(target)) {
				itSources.remove();
				itTargets.remove();
				itActivities.remove();
			}
		}
	}

	/**
	 * Iterator to iterate over the outgoing edges of a state.
	 * 
	 * @author sander
	 *
	 */
	public final class EdgeIterableOutgoing {
		private int source;
		private int next = 0;
		private int current = 0;

		private EdgeIterableOutgoing(int source) {
			if (source >= 0) {
				reset(source);
			}
		}

		/**
		 * Sets the iterator to the beginning of the edges of source. If source
		 * is -1, the iterator will be empty.
		 * 
		 * @param source
		 */
		public void reset(int source) {
			this.source = source;
			if (source == -1) {
				next = sources.size();
				current = sources.size() - 1;
			} else {
				int from = sources.binarySearch(source);
				if (from < 0) {
					return;
				}
				while (from >= 0 && sources.get(from) == source) {
					from--;
				}
				next = from + 1;
				current = -1;
			}
		}

		public boolean hasNext() {
			return next < getSources().size() && getSources().get(next) == source;
		}

		public int getTarget() {
			return targets.get(current);
		}

		public short getActivity() {
			return activities.get(current);
		}

		public int nextEdge() {
			current = next;
			next++;
			assert(current < activities.size());
			return current;
		}

		public void remove() {
			sources.removeAt(current);
			targets.removeAt(current);
			activities.removeAt(current);
			next--;
			current--;
		}

		public int nextTarget() {
			current = next;
			next++;
			return getTarget();
		}

		public short nextActivity() {
			current = next;
			next++;
			return getActivity();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#toString()
	 */
	@Override
	public String toString() {
		return toDot();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#toDot()
	 */
	@Override
	public String toDot() {
		StringBuilder b = new StringBuilder("digraph Automaton {\n");
		b.append("  rankdir = LR;\n");
		b.append("  node [shape=circle]");

		//draw the initial state
		b.append("  initial [shape=plaintext,label=\"\"];\n");
		b.append("  initial -> 0;\n");

		//draw non-initial accepting states
		int state = accepting.nextSetBit(0);
		while (state >= 0) {
			b.append("  ");
			b.append(state);
			b.append(" [shape=doublecircle];\n");

			state = accepting.nextSetBit(state + 1);
		}

		//draw transitions
		for (int i = 0; i < sources.size(); i++) {
			b.append("  ");
			b.append(sources.get(i));
			b.append(" -> ");
			b.append(targets.get(i));
			b.append(" [label=\"");
			b.append(activities.get(i));
			b.append("\"];\n");
		}

		return b.append("}\n").toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#getOutgoingEdgesIterator(int)
	 */
	@Override
	public EdgeIterableOutgoing getOutgoingEdgesIterator(int state) {
		return new EdgeIterableOutgoing(state);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#getNumberOfStates()
	 */
	@Override
	public int getNumberOfStates() {
		return maxState + 1;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#getTargets()
	 */
	@Override
	public TIntList getTargets() {
		return TCollections.unmodifiableList(targets);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#getSources()
	 */
	@Override
	public TIntList getSources() {
		return TCollections.unmodifiableList(sources);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#getActivities()
	 */
	public TShortList getActivities() {
		return TCollections.unmodifiableList(activities);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.projectedrecallandprecision.helperclasses.automata.
	 * Automaton2I#step(int, short)
	 */
	@Override
	public int step(int state, short activity) {
		if (state == -1) {
			return -1;
		}
		int from = binarySearch(state, activity);
		if (from >= 0) {
			return targets.get(from);
		}
		return -1;
	}
}
