package org.processmining.projectedrecallandprecision.helperclasses.automata;

import java.util.BitSet;

import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList.EdgeIterableOutgoing;

import gnu.trove.list.TIntList;

public class DFAOperations {

	/**
	 * Adds the trace to the automaton in a branching manner, such that the
	 * automaton accepts the trace.
	 * 
	 * @param automaton
	 * @param trace
	 * @param canceller
	 * @return The last state
	 */
	public static int addTrace(DFA automaton, short[] trace, ProMCanceller canceller) {
		int state = automaton.getInitialState();
		for (int i = 0; i < trace.length; i++) {
			short activity = trace[i];
			state = automaton.addEdge(state, activity);

			if (canceller.isCancelled()) {
				return -1;
			}
		}
		automaton.setAccepting(state, true);
		return state;
	}

	/**
	 * 
	 * @param automaton
	 * @return a bitset with the states that are reachable in the automaton.
	 */
	public static BitSet getReachableStates(DFA automaton) {
		BitSet result = new BitSet(automaton.getNumberOfStates());
		BitSet toProcess = new BitSet(automaton.getNumberOfStates());
		toProcess.set(automaton.getInitialState());
		EdgeIterableOutgoing it = automaton.getOutgoingEdgesIterator(automaton.getInitialState());

		int state = toProcess.nextSetBit(0);
		while (state >= 0) {
			result.set(state);
			toProcess.set(state, false);

			it.reset(state);
			while (it.hasNext()) {
				int target = it.nextTarget();
				if (!result.get(target)) {
					toProcess.set(target);
				}
			}

			state = toProcess.nextSetBit(0);
		}

		return result;
	}

	/**
	 * 
	 * @param automaton
	 * @return The states from which an accepting state is reachable.
	 */
	public static BitSet getLiveStates(DFA automaton) {
		BitSet liveStates = automaton.getAcceptingStates();
		BitSet worklist = automaton.getAcceptingStates();
		BitSet nextWorklist = null;
		TIntList sources = automaton.getSources();
		TIntList targets = automaton.getTargets();

		while (!worklist.isEmpty()) {
			nextWorklist = new BitSet(automaton.getNumberOfStates());
			for (int i = 0; i < targets.size(); i++) {
				int target = targets.get(i);
				if (worklist.get(target)) {
					int source = sources.get(i);
					if (!liveStates.get(source)) {
						nextWorklist.set(source);
						liveStates.set(source);
					}
				}
			}
			worklist = nextWorklist;
		}

		return liveStates;
	}

	/**
	 * 
	 * @param s
	 * @return whether the automaton accepts the trace. Might not be very
	 *         efficient; use RunAutomaton2 instead.
	 */
	public static boolean accepts(DFA automaton, short[] trace) {
		int state = walk(automaton, trace);
		return state != -1 && automaton.isAccepting(state);
	}

	/**
	 * Performs a walk through the automaton. Might not be very efficient.
	 * 
	 * @param automaton
	 * @param trace
	 * @return the state after the walk, or -1 if the walk is not possible in
	 *         the automaton.
	 */
	public static int walk(DFA automaton, short[] trace) {
		int state = automaton.getInitialState();
		for (int i = 0; i < trace.length; i++) {
			short activity = trace[i];
			state = automaton.step(state, activity);
		}
		return state;
	}
}