package org.processmining.projectedrecallandprecision.helperclasses.automata.operations;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList.EdgeIterableOutgoing;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAOperations;

public class Automaton2Minimise {
	/**
	 * Minimise the automaton with the Hopcroft algorithm
	 * 
	 * @param automaton
	 * @param canceller
	 */
	public static void minimise(DFA automaton, ProMCanceller canceller) {
		removeDeadStates(automaton);

		int numberOfStates = automaton.getNumberOfStates();

		boolean[][] D = new boolean[numberOfStates][numberOfStates];
		List<List<Set<Point>>> S = new ArrayList<>();

		for (int i = 0; i < numberOfStates; i++) {
			List<Set<Point>> innerList = new ArrayList<>();

			for (int j = 0; j < numberOfStates; j++) {
				Arrays.fill(D[i], false);
				innerList.add(new HashSet<Point>());
			}
			S.add(innerList);

			if (canceller.isCancelled()) {
				return;
			}
		}

		for (int i = 0; i < numberOfStates; i++) {
			for (int j = i + 1; j < numberOfStates; j++) {
				if (automaton.isAccepting(i) != automaton.isAccepting(j)) {
					D[i][j] = true;
				}
			}

			if (canceller.isCancelled()) {
				return;
			}
		}

		EdgeIterableOutgoing itI = automaton.getOutgoingEdgesIterator(-1);
		EdgeIterableOutgoing itJ = automaton.getOutgoingEdgesIterator(-1);
		for (int i = 0; i < numberOfStates; i++) {
			for (int j = i + 1; j < numberOfStates; j++) {
				// only consider pairs that are as of yet indistinguishable
				if (D[i][j]) {
					continue;
				}

				// helps emulate "for any"
				boolean distinguished = false;
				{
					itI.reset(i);
					itJ.reset(j);
					while (itI.hasNext() && itJ.hasNext()) {
						int targetI = itI.nextTarget();
						int targetJ = itJ.nextTarget();

						// if the targets are distinguishable, so are i and j
						if (D[targetI][targetJ] || D[targetJ][targetI]) {
							dist(S, D, i, j);
							distinguished = true;
							break;
						}

						//if the activities are not the same (they are sorted), then i and j are distinguishable
						if (itI.getActivity() != itJ.getActivity()) {
							dist(S, D, i, j);
							distinguished = true;
							break;
						}
					}

					//if the number of activities is not the same, then i and j are distinguishable
					if (itI.hasNext() || itJ.hasNext()) {
						dist(S, D, i, j);
						distinguished = true;
					}
				}

				if (!distinguished) {
					// qm and qn are indistinguishable
					itI.reset(i);
					itJ.reset(j);
					while (itI.hasNext() && itJ.hasNext()) {
						int m = itI.nextTarget();
						int n = itJ.nextTarget();

						if (m < n && !(i == m && j == n)) {
							S.get(m).get(n).add(new Point(i, j));
						} else if (m > n && !(i == n && j == m)) {
							S.get(n).get(m).add(new Point(i, j));
						}
					}
				}
			}

			if (canceller.isCancelled()) {
				return;
			}
		}

		// merge states together by smallest equivalent
		BitSet reachable = DFAOperations.getReachableStates(automaton);
		for (int i = 0; i < D.length; i++) {
			for (int j = i + 1; j < D.length; j++) {
				if (reachable.get(i) && reachable.get(j) && !D[i][j]) {
					//merge i and j
					automaton.removeStateAndRedirectEdgesTo(j, i);
				}
			}

			if (canceller.isCancelled()) {
				return;
			}
		}
	}

	private static void dist(List<List<Set<Point>>> S, boolean[][] D, int i, int j) {
		_dist(S, D, new Point(i, j), new HashSet<Point>());
	}

	private static void _dist(List<List<Set<Point>>> S, boolean[][] D, Point point, HashSet<Point> visited) {
		if (visited.contains(point)) {
			return;
		}

		int i = point.x, j = point.y;
		D[i][j] = true;
		visited.add(point);
		for (Point pair : S.get(i).get(j)) {
			_dist(S, D, pair, visited);
		}
	}

	/**
	 * Remove the states from which no accepting state can be reached.
	 * 
	 * @param automaton
	 */
	public static void removeDeadStates(DFA automaton) {
		BitSet liveStates = DFAOperations.getLiveStates(automaton);
		automaton.removeStatesExcept(liveStates);
	}
}
