package org.processmining.projectedrecallandprecision.helperclasses.automata.operations;

import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList.EdgeIterableOutgoing;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.StatePairSet.StatePairIterator;

import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.TShortObjectMap;
import gnu.trove.map.custom_hash.TObjectIntCustomHashMap;
import gnu.trove.map.hash.TShortObjectHashMap;
import gnu.trove.procedure.TShortObjectProcedure;
import gnu.trove.strategy.HashingStrategy;

public class Automaton2Shuffle {

	/**
	 * Returns a new automaton that accepts the shuffle product of automatonA
	 * and automatonB. Does not modify the input automata.
	 * 
	 * @param automatonA
	 * @param automatonB
	 * @param canceller
	 * @return
	 */
	public static DFA shuffle(List<DFA> automata, ProMCanceller canceller) {
		Iterator<DFA> it = automata.iterator();
		if (!it.hasNext()) {
			return null;
		}

		DFA result = it.next();

		while (it.hasNext()) {
			result = shuffle(result, it.next(), canceller);

			if (result == null || canceller.isCancelled()) {
				return null;
			}
		}

		return result;
	}

	public static DFA shuffle(DFA automatonA, DFA automatonB, ProMCanceller canceller) {
		final DFA shuffle = new DFAImplEdgeList();

		//no way around it here: we need NFA-like structures.
		final ArrayDeque<int[]> worklist = new ArrayDeque<>();
		final TObjectIntMap<int[]> statePair2shuffleState = new TObjectIntCustomHashMap<>(new HashingStrategy<int[]>() {
			private static final long serialVersionUID = -4818184406522740901L;

			public int computeHashCode(int[] object) {
				return Arrays.hashCode(object);
			}

			public boolean equals(int[] o1, int[] o2) {
				return Arrays.equals(o1, o2);
			}
		}, 10, 0.5f, -1);
		{
			int[] statePair = StatePairSet.create(automatonA.getInitialState(), automatonB.getInitialState());
			worklist.add(statePair);
			statePair2shuffleState.put(statePair, shuffle.getInitialState());
		}

		EdgeIterableOutgoing itA = automatonA.getOutgoingEdgesIterator(automatonA.getInitialState());
		EdgeIterableOutgoing itB = automatonB.getOutgoingEdgesIterator(automatonB.getInitialState());
		StatePairIterator itStates = new StatePairSet.StatePairIterator();

		while (!worklist.isEmpty()) {
			int[] sourcePairSet = worklist.removeFirst();
			final int shuffleSourceState = statePair2shuffleState.get(sourcePairSet);

			//set the state accepting if necessary
			shuffle.setAccepting(shuffleSourceState,
					StatePairSet.aPairIsBothAccepting(automatonA, automatonB, sourcePairSet));

			//gather the next states
			TShortObjectMap<int[]> nextStates = new TShortObjectHashMap<>(10, 0.5f, (short) -1);

			for (itStates.reset(sourcePairSet); itStates.hasNext();) {
				itStates.next();

				//walk through the edges of nodeA; add a possibility for all outgoing edges.
				int sourceStateA = itStates.getNodeA();
				int sourceStateB = itStates.getNodeB();

				//add the possible states going from this stateA in this statePair
				{
					for (itA.reset(sourceStateA); itA.hasNext();) {
						itA.nextEdge();

						int targetStateA = itA.getTarget();
						int targetStateB = sourceStateB;
						short activity = itA.getActivity();

						if (nextStates.containsKey(activity)) {
							nextStates.put(activity,
									StatePairSet.add(nextStates.get(activity), targetStateA, targetStateB));
						} else {
							nextStates.put(activity, StatePairSet.create(targetStateA, targetStateB));
						}
					}
				}
				
				//add the possible states going from this stateB in this statePair
				{
					for (itB.reset(sourceStateB); itB.hasNext();) {
						itB.nextEdge();

						int targetStateA = sourceStateA;
						int targetStateB = itB.getTarget();
						short activity = itB.getActivity();

						if (nextStates.containsKey(activity)) {
							nextStates.put(activity,
									StatePairSet.add(nextStates.get(activity), targetStateA, targetStateB));
						} else {
							nextStates.put(activity, StatePairSet.create(targetStateA, targetStateB));
						}
					}
				}
			}

			nextStates.forEachEntry(new TShortObjectProcedure<int[]>() {
				public boolean execute(short activity, int[] statePairSetTarget) {
					int unionTargetState = statePair2shuffleState.get(statePairSetTarget);
					if (unionTargetState == statePair2shuffleState.getNoEntryValue()) {
						//this state pair did not exist yet
						unionTargetState = shuffle.addEdge(shuffleSourceState, activity);
						statePair2shuffleState.put(statePairSetTarget, unionTargetState);

						worklist.add(statePairSetTarget);
					} else {
						//this state pair did not exist yet
						shuffle.addEdge(shuffleSourceState, activity, unionTargetState);
					}
					return true;
				}
			});
		}

		return shuffle;
	}
}
