package org.processmining.projectedrecallandprecision.plugins;

import java.awt.Desktop;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;

import javax.swing.Box;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextArea;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.apache.commons.lang.StringUtils;
import org.deckfour.xes.model.XLog;
import org.processmining.plugins.InductiveMiner.ClassifierChooser;
import org.processmining.projectedrecallandprecision.framework.CompareParameters;
import org.processmining.projectedrecallandprecision.framework.CompareParameters.RecallName;

import com.fluxicon.slickerbox.factory.SlickerFactory;

public class CompareParametersDialog extends JPanel {

	private static final long serialVersionUID = 7693870370139578439L;
	private final CompareParameters parameters;
	private final JLabel subsetLabel;
	private final JSlider subsetSlider;
	private final JLabel subsetValue;
	private final JCheckBox recall;
	private final JCheckBox precision;
	private final JLabel doiLabel;
	private final JLabel doiValue;

	public final String doi = "http://dx.doi.org/10.1007/s10270-016-0545-x";

	public CompareParametersDialog(XLog log, int numberOfActivities, RecallName recallName) {
		parameters = new CompareParameters(recallName);

		SlickerFactory factory = SlickerFactory.instance();

		int gridy = 1;

		setLayout(new GridBagLayout());

		//subset size
		subsetLabel = factory.createLabel("Size of projection");
		{
			GridBagConstraints cNoiseLabel = new GridBagConstraints();
			cNoiseLabel.gridx = 0;
			cNoiseLabel.gridy = gridy;
			cNoiseLabel.anchor = GridBagConstraints.WEST;
			add(subsetLabel, cNoiseLabel);
		}

		subsetSlider = factory.createSlider(SwingConstants.HORIZONTAL);
		{
			subsetSlider.setMinimum(1);
			subsetSlider.setMaximum(numberOfActivities);
			GridBagConstraints cNoiseSlider = new GridBagConstraints();
			cNoiseSlider.gridx = 1;
			cNoiseSlider.gridy = gridy;
			cNoiseSlider.fill = GridBagConstraints.HORIZONTAL;
			add(subsetSlider, cNoiseSlider);
		}

		subsetValue = factory.createLabel("   ");
		{
			GridBagConstraints cNoiseValue = new GridBagConstraints();
			cNoiseValue.gridx = 2;
			cNoiseValue.gridy = gridy;
			add(subsetValue, cNoiseValue);
		}

		gridy++;

		final JTextArea subsetExplanation = new JTextArea("Sets the size of the projection.\n"
				+ "A low size is faster and more forgiving.\n" + "A high size is more accurate.\n"
				+ "If set to maximum, the answer will be equivalent to language inclusion.");
		{
			subsetExplanation.setEditable(false);
			subsetExplanation.setEnabled(false);
			subsetExplanation.setFont(subsetValue.getFont());
			subsetExplanation.setOpaque(false);
			subsetExplanation.setForeground(subsetValue.getForeground());
			subsetExplanation.setDisabledTextColor(subsetValue.getForeground());
			GridBagConstraints cNoiseExplanation = new GridBagConstraints();
			cNoiseExplanation.gridx = 1;
			cNoiseExplanation.gridy = gridy;
			cNoiseExplanation.gridwidth = 3;
			cNoiseExplanation.anchor = GridBagConstraints.WEST;
			add(subsetExplanation, cNoiseExplanation);
		}

		gridy++;

		//spacer
		{
			JLabel spacer = factory.createLabel(" ");
			GridBagConstraints cSpacer = new GridBagConstraints();
			cSpacer.gridx = 0;
			cSpacer.gridy = gridy;
			cSpacer.anchor = GridBagConstraints.WEST;
			add(spacer, cSpacer);
		}

		gridy++;

		final ClassifierChooser classifiers;
		if (log != null) {
			//classifiers
			{
				final JLabel classifierLabel = factory.createLabel("Event classifier");
				GridBagConstraints cClassifierLabel = new GridBagConstraints();
				cClassifierLabel.gridx = 0;
				cClassifierLabel.gridy = gridy;
				cClassifierLabel.weightx = 0.4;
				cClassifierLabel.anchor = GridBagConstraints.NORTHWEST;
				add(classifierLabel, cClassifierLabel);
			}

			classifiers = new ClassifierChooser(log);
			{
				GridBagConstraints cClassifiers = new GridBagConstraints();
				cClassifiers.gridx = 1;
				cClassifiers.gridy = gridy;
				cClassifiers.anchor = GridBagConstraints.NORTHWEST;
				cClassifiers.fill = GridBagConstraints.HORIZONTAL;
				cClassifiers.weightx = 0.6;
				add(classifiers, cClassifiers);
			}
			gridy++;

			//spacer
			{
				JLabel spacer = factory.createLabel(" ");
				GridBagConstraints cSpacer = new GridBagConstraints();
				cSpacer.gridx = 0;
				cSpacer.gridy = gridy;
				cSpacer.anchor = GridBagConstraints.WEST;
				add(spacer, cSpacer);
			}
		} else {
			classifiers = null;
		}

		gridy++;

		//compute recall
		{
			JLabel recallLabel = factory.createLabel("Compute " + parameters.getRecallName());
			GridBagConstraints cRecallLabel = new GridBagConstraints();
			cRecallLabel.gridx = 0;
			cRecallLabel.gridy = gridy;
			cRecallLabel.weightx = 0.4;
			cRecallLabel.anchor = GridBagConstraints.NORTHWEST;
			add(recallLabel, cRecallLabel);
		}

		{
			recall = factory.createCheckBox("", parameters.isComputeRecallFitness());
			GridBagConstraints cRecall = new GridBagConstraints();
			cRecall.gridx = 1;
			cRecall.gridy = gridy;
			cRecall.anchor = GridBagConstraints.NORTHWEST;
			cRecall.weightx = 0.6;
			add(recall, cRecall);
		}

		gridy++;

		//compute precision
		{
			JLabel precisionLabel = factory.createLabel("Compute precision");
			GridBagConstraints cPrecisionLabel = new GridBagConstraints();
			cPrecisionLabel.gridx = 0;
			cPrecisionLabel.gridy = gridy;
			cPrecisionLabel.weightx = 0.4;
			cPrecisionLabel.anchor = GridBagConstraints.NORTHWEST;
			add(precisionLabel, cPrecisionLabel);
		}

		{
			precision = factory.createCheckBox("", parameters.isComputePrecision());
			GridBagConstraints cRecall = new GridBagConstraints();
			cRecall.gridx = 1;
			cRecall.gridy = gridy;
			cRecall.anchor = GridBagConstraints.NORTHWEST;
			cRecall.weightx = 0.6;
			add(precision, cRecall);
		}

		gridy++;

		//spacer
		{
			JLabel spacer = factory.createLabel(" ");
			GridBagConstraints cSpacer = new GridBagConstraints();
			cSpacer.gridx = 0;
			cSpacer.gridy = gridy;
			cSpacer.anchor = GridBagConstraints.WEST;
			add(spacer, cSpacer);
		}

		gridy++;

		//hints
		{

			{
				final JLabel hintsLabel = factory.createLabel("NB.");
				GridBagConstraints cHintsLabel = new GridBagConstraints();
				cHintsLabel.gridx = 0;
				cHintsLabel.gridy = gridy;
				cHintsLabel.weightx = 0.4;
				cHintsLabel.anchor = GridBagConstraints.NORTHWEST;
				add(hintsLabel, cHintsLabel);
			}

			{
				JTextArea hints = new JTextArea("Only models that are weakly sound and bounded are supported.\n"
						+ "Applying to a weakly unsound or unbounded model will give an error.\n"
						+ "\nIf possible, use process trees instead of Petri nets for the highest speed.");
				hints.setEditable(false);
				hints.setEnabled(false);
				hints.setFont(subsetValue.getFont());
				hints.setOpaque(false);
				hints.setForeground(subsetValue.getForeground());
				hints.setDisabledTextColor(subsetValue.getForeground());
				GridBagConstraints cHints = new GridBagConstraints();
				cHints.gridx = 1;
				cHints.gridy = gridy;
				cHints.gridwidth = 3;
				cHints.anchor = GridBagConstraints.WEST;
				add(hints, cHints);
			}
		}

		gridy++;

		//spacer
		{
			JLabel spacer = factory.createLabel(" ");
			GridBagConstraints cSpacer = new GridBagConstraints();
			cSpacer.gridx = 0;
			cSpacer.gridy = gridy;
			cSpacer.anchor = GridBagConstraints.WEST;
			add(spacer, cSpacer);
		}

		gridy++;

		//doi
		{
			doiLabel = factory.createLabel("More information");
			GridBagConstraints cDoiLabel = new GridBagConstraints();
			cDoiLabel.gridx = 0;
			cDoiLabel.gridy = gridy;
			cDoiLabel.weightx = 0.4;
			cDoiLabel.anchor = GridBagConstraints.NORTHWEST;
			add(doiLabel, cDoiLabel);
		}

		{
			doiValue = factory.createLabel("doi doi");
			GridBagConstraints cDoiValue = new GridBagConstraints();
			cDoiValue.gridx = 1;
			cDoiValue.gridy = gridy;
			cDoiValue.anchor = GridBagConstraints.NORTHWEST;
			cDoiValue.weightx = 0.6;
			add(doiValue, cDoiValue);
		}

		gridy++;

		{
			GridBagConstraints gbcFiller = new GridBagConstraints();
			gbcFiller.weighty = 1.0;
			gbcFiller.gridy = gridy;
			gbcFiller.fill = GridBagConstraints.BOTH;
			add(Box.createGlue(), gbcFiller);
		}

		//controller

		subsetSlider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent arg0) {
				parameters.setK(subsetSlider.getValue());
				subsetValue.setText(StringUtils.leftPad(parameters.getK() + "", 3, ' '));
			}
		});
		subsetSlider.setValue(parameters.getK());
		subsetValue.setText(StringUtils.leftPad(parameters.getK() + "", 3, ' '));

		recall.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				parameters.setComputeRecallFitness(recall.isSelected());
			}
		});

		precision.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				parameters.setComputePrecision(precision.isSelected());
			}
		});

		if (log != null) {
			classifiers.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent arg0) {
					parameters.setClassifier(classifiers.getSelectedClassifier());
				}
			});
			parameters.setClassifier(classifiers.getSelectedClassifier());
		}

		doiValue.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				openWebPage(doi);
			}
		});
		doiValue.setText(doi);
	}

	public CompareParameters getMiningParameters() {
		return parameters;
	}

	public static void openWebPage(String url) {
		if (Desktop.isDesktopSupported()) {
			Desktop desktop = Desktop.getDesktop();
			try {
				desktop.browse(new URI(url));
			} catch (IOException | URISyntaxException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		} else {
			Runtime runtime = Runtime.getRuntime();
			try {
				runtime.exec("xdg-open " + url);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
}